#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil import testing_reqs
from coremltools.converters.mil.testing_reqs import *
from coremltools.converters.mil.mil import Program, Function, get_new_symbol

from .testing_utils import UNK_SYM, run_compare_builder

backends = testing_reqs.backends


class TestNormalizationBatchNorm:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x_val = np.array(
            [
                [
                    [[-16.0, 13.0], [11.0, -16.0]],
                    [[13.0, -15.0], [13.0, 9.0]],
                    [[-9.0, -4.0], [-6.0, 3.0]],
                ]
            ],
            dtype=np.float32,
        )
        mean_val = np.array([9.0, 6.0, 3.0], dtype=np.float32)
        variance_val = np.array([6.0, 1.0, 7.0], dtype=np.float32)
        gamma_val = np.array([1.0, 1.0, 1.0], dtype=np.float32)
        beta_val = np.array([1.0, 3.0, 0.0], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [
                mb.batch_norm(x=x, mean=mean_val, variance=variance_val),
                mb.batch_norm(
                    x=x,
                    mean=mean_val,
                    variance=variance_val,
                    gamma=gamma_val,
                    beta=beta_val,
                    epsilon=1e-4,
                ),
            ]

        expected_output_types = [
            (1, 3, 2, 2, types.fp32),
            (1, 3, 2, 2, types.fp32),
        ]
        expected_outputs = [
            np.array(
                [
                    [
                        [[-10.206199, 1.6329918], [0.8164959, -10.206199]],
                        [[6.999965, -20.999895], [6.999965, 2.9999852]],
                        [[-4.53557, -2.6457493], [-3.4016776, 0.0]],
                    ]
                ],
                dtype=np.float32,
            ),
            np.array(
                [
                    [
                        [[-9.206122, 2.6329796], [1.8164899, -9.206122]],
                        [[9.99965, -17.998951], [9.99965, 5.9998503]],
                        [[-4.535541, -2.6457324], [-3.4016557, 0.0]],
                    ]
                ],
                dtype=np.float32,
            ),
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )


class TestNormalizationInstanceNorm:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x_val = np.array(
            [
                [
                    [[-16.0, 13.0], [11.0, 16.0]],
                    [[13.0, 15.0], [13.0, 9.0]],
                    [[-9.0, 4.0], [-6.0, 3.0]],
                ],

                [
                    [[-5.0, 1.0], [12.0, 3.0]],
                    [[0.0, 9.0], [2.0, -8.0]],
                    [[2.0, 5.0], [10.0, 0.0]],

                ]
            ],
            dtype=np.float32,
        )
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return mb.instance_norm(x=x, epsilon=1e-2)

        expected_output_types = [(2, 3, 2, 2, types.fp32)]
        expected_outputs = [
            np.array(
                [
                    [
                        [[-1.71524656, 0.54576027], [0.38982874, 0.77965748]],
                        [[0.22917463, 1.14587319], [0.22917463, -1.60422242]],
                        [[-1.2470212, 1.06887531], [-0.71258354, 0.89072943]],
                    ],

                    [
                        [[-1.27070526, -0.28693344], [1.51664821, 0.04099049]],
                        [[-0.12380638, 1.36187018], [0.20634397, -1.44440776]],
                        [[-0.59714057, 0.19904686], [1.5260259, -1.12793219]],
                    ]
                ],
                dtype=np.float32,
            )
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke_with_gamma_and_beta(self, use_cpu_only, backend):
        x_val = np.array(
            [
                [
                    [[-16.0, 13.0], [11.0, 16.0]],
                    [[13.0, 15.0], [13.0, 9.0]],
                    [[-9.0, 4.0], [-6.0, 3.0]],
                ],

                [
                    [[-5.0, 1.0], [12.0, 3.0]],
                    [[0.0, 9.0], [2.0, -8.0]],
                    [[2.0, 5.0], [10.0, 0.0]],

                ]
            ],
            dtype=np.float32,
        )
        gamma_val = np.array([-9.0, 3.2, 1.3], dtype=np.float32)
        beta_val = np.array([-0.8, 3.4, 1.2], dtype=np.float32)

        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return mb.instance_norm(x=x, gamma=gamma_val, beta=beta_val, epsilon=1e-2)

        expected_output_types = [(2, 3, 2, 2, types.fp32)]
        expected_outputs = [
            np.array(
                [
                    [
                        [[14.63721807, -5.71184211], [-4.30845865, -7.8169173]],
                        [[4.1333588, 7.06679399], [4.1333588, -1.73351158]],
                        [[-0.42112757, 2.58953791], [0.27364139, 2.35794826]],
                    ],

                    [
                        [[10.6363473, 1.782401], [-14.44983388, -1.16891443]],
                        [[3.00381959, 7.75798456], [4.06030069, -1.22210484]],
                        [[0.42371726, 1.45876091], [3.18383368, -0.26631185]],
                    ]
                ],
                dtype=np.float32,
            )
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.skipif(not testing_reqs._HAS_TORCH, reason="PyTorch not found.")
    @pytest.mark.parametrize(
        "rank, use_cpu_only, backend, epsilon",
        itertools.product([3, 4], [True, False], backends, [1e-3, 1e-5, 1e-10]),
    )
    def test_builder_to_backend_stress(self, rank, use_cpu_only, backend, epsilon):
        shape = np.random.randint(low=2, high=6, size=rank)
        x_val = random_gen(shape=shape, rand_min=-100.0, rand_max=100.0)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return mb.instance_norm(x=x, epsilon=epsilon)

        layer = torch.nn.InstanceNorm2d if rank == 4 else torch.nn.InstanceNorm1d
        torch_op = layer(num_features=shape[1], eps=epsilon)
        expected_outputs = [torch_op(torch.as_tensor(x_val)).numpy()]
        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
            atol=1e-3,
            rtol=1e-4,
            also_compare_shapes=True
        )


class TestNormalizationL2Norm:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x_val = np.array([[[1.0, -7.0], [5.0, -6.0], [-3.0, -5.0]]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [mb.l2_norm(x=x, epsilon=1e-10)]

        expected_output_types = [(1, 3, 2, types.fp32)]
        expected_outputs = [
            np.array(
                [
                    [
                        [0.08304548, -0.58131838],
                        [0.41522741, -0.4982729],
                        [-0.24913645, -0.41522741],
                    ]
                ],
                dtype=np.float32,
            )
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend, rank", itertools.product([True, False], backends, [3, 4, 5])
    )
    def test_builder_to_backend_stress(self, use_cpu_only, backend, rank):
        shape = np.random.randint(low=2, high=6, size=rank)
        x_val = random_gen(shape=shape, rand_min=-10.0, rand_max=10.0)
        input_placeholders = {"x": mb.placeholder(shape=shape)}
        input_values = {"x": x_val}

        def build(x):
            return [mb.l2_norm(x=x, epsilon=1e-12)]

        # compute for the answer
        batch_dims = rank - 3
        if batch_dims == 0:
            norm = la.norm(x_val)
            output = x_val/norm
        else:
            batch_dim_prod = np.prod(shape[:batch_dims])
            reshape_x_val = np.reshape(x_val,(batch_dim_prod,-1))
            norm = la.norm(reshape_x_val, axis=1, keepdims=True)
            output = reshape_x_val/norm
            output = np.reshape(output, shape)

        expected_output_types = [list(output.shape) + [types.fp32]]
        expected_outputs = [
            output
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )


class TestNormalizationLayerNorm:

    @staticmethod
    def _keras_layer_norm( x, axes, epsilon):
            layer = tf.keras.layers.LayerNormalization(axis=axes, epsilon=epsilon)
            data = tf.constant(x, dtype=tf.float32)
            output = layer(data)
            return output.numpy()

    @staticmethod
    def _np_layer_norm(x, axes, gamma=None, beta=None, epsilon=1e-5):
            rank = len(x.shape)
            axes = [axis + rank if axis < 0 else axis for axis in axes]
            normalized_shape = [x.shape[i] if i in axes else 1 for i in range(rank)]
            gamma = np.ones(shape=normalized_shape) if gamma is None else np.reshape(gamma, normalized_shape)
            beta = np.zeros(shape=normalized_shape) if beta is None else np.reshape(beta, normalized_shape)
            num = x - np.mean(x, axis=tuple(axes), keepdims=True)
            dem = np.sqrt(
                np.sum(np.square(num), axis=tuple(axes), keepdims=True)
                / np.prod(normalized_shape)
                + epsilon
            )
            return num / dem * gamma + beta

    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x_val = np.array([[[1.0, -7.0], [5.0, -6.0], [-3.0, -5.0]]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [
                # V2->V1 lowering (op_mappings.py): if branch
                mb.layer_norm(x=x, axes=[2], epsilon=1e-4),
                # V2->V1 lowering (op_mappings.py): else branch
                mb.layer_norm(x=x, axes=[-2, -1], epsilon=1e-4),
            ]

        expected_output_types = [(1, 3, 2, types.fp32), (1, 3, 2, types.fp32)]
        expected_outputs = [
            np.array(
                [
                    [
                        [ 0.9999969,  -0.9999969 ],
                        [ 0.99999833, -0.99999833],
                        [ 0.99995005, -0.99995005],
                    ]
                ],
                dtype=np.float32,
            ),
            np.array(
                [
                    [
                        [ 0.82687193, -1.06312108],
                        [ 1.77186835, -0.82687193],
                        [-0.11812456, -0.59062278],
                    ]
                ],
                dtype=np.float32,
            ),
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke_with_dynamic_shape(self, use_cpu_only, backend):
        x_val = np.array([[[1.0, -7.0], [5.0, -6.0], [-3.0, -5.0]]], dtype=np.float32)
        shape = (get_new_symbol(), get_new_symbol(), 2)
        input_placeholders = {"x": mb.placeholder(shape=shape)}
        input_values = {"x": x_val}

        def build(x):
            return [
                mb.layer_norm(x=x, axes=[2], epsilon=1e-4),
            ]

        expected_output_types = [(UNK_SYM, UNK_SYM, 2, types.fp32)]
        expected_outputs = [
            np.array(
                [
                    [
                        [ 0.9999969,  -0.9999969 ],
                        [ 0.99999833, -0.99999833],
                        [ 0.99995005, -0.99995005],
                    ]
                ],
                dtype=np.float32,
            ),
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend, rank_and_axes, epsilon, provides_gamma_beta",
             itertools.product([True, False], backends,
            [[3,[0,2]], [3,[-2]], [4,[0,1,3]], [5,[0,4]], [5,[-5,-4,-3,-2,-1]]
            ],
            [0.0001, 0.01],
            [True, False]),
        )
    def test_builder_to_backend_stress_numpy(self, use_cpu_only, backend, rank_and_axes, epsilon, provides_gamma_beta):
        rank, axes = rank_and_axes
        shape = np.random.randint(low=2, high=6, size=rank)
        x_val = random_gen(shape=shape, rand_min=-100.0, rand_max=100.0)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        gamma, beta = None, None

        if provides_gamma_beta:
            positive_axes = [axis+rank if axis <0 else axis for axis in axes]
            normalized_shape = [shape[i] for i in range(rank) if i in positive_axes]
            gamma = random_gen(shape=normalized_shape, rand_min=-100, rand_max=100)
            beta = random_gen(shape=normalized_shape, rand_min=-100, rand_max=100)

        def build(x):
            return [
                mb.layer_norm(x=x, axes=axes, epsilon=epsilon, gamma=gamma, beta=beta)
            ]

        output = TestNormalizationLayerNorm._np_layer_norm(x=x_val, axes=axes, epsilon=epsilon, gamma=gamma, beta=beta)
        expected_output_types = [tuple(output.shape) + (types.fp32,)]
        expected_outputs = [
            output
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
            atol=1e-3,
            rtol=1e-4,
        )

    @pytest.mark.skipif(not testing_reqs._HAS_TF_2, reason="Tensorflow not found.")
    @pytest.mark.parametrize(
        "use_cpu_only, backend, rank_and_axes, epsilon",
             itertools.product([True, False], backends,
            [[3,[0,2]], [3,[-2]], [4,[0,1,3]], [5,[0,4]], [5,[-5,-4,-3,-2,-1]]
            ],
            [0.0001, 0.01]),
        )
    def test_builder_to_backend_stress_keras(self, use_cpu_only, backend, rank_and_axes, epsilon):
        rank, axes = rank_and_axes
        shape = np.random.randint(low=2, high=6, size=rank)
        x_val = random_gen(shape=shape, rand_min=-100.0, rand_max=100.0)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [
                mb.layer_norm(x=x, axes=axes, epsilon=epsilon)
            ]

        output = TestNormalizationLayerNorm._keras_layer_norm(x=x_val, axes=axes, epsilon=epsilon)
        expected_output_types = [tuple(output.shape) + (types.fp32,)]
        expected_outputs = [
            output
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.parametrize("rank_and_axes, epsilon",
        itertools.product(
            [[3,[0,2]], [3,[-2,-1]], [4,[0,1,2,3]], [5,[0,2,-1]], [5,[-5,-4,-3,-2,-1]]],
            [0.0001, 0.01],
        ),
    )
    def test_builder_eval_stress(self, rank_and_axes, epsilon):
        rank, axes = rank_and_axes
        shape = np.random.randint(low=2, high=6, size=rank)
        x_val = random_gen(shape=shape, rand_min=-100.0, rand_max=100.0)
        positive_axes = [axis+rank if axis <0 else axis for axis in axes]
        normalized_shape = [shape[i] for i in range(rank) if i in positive_axes]
        gamma_val = random_gen(shape=normalized_shape, rand_min=-100, rand_max=100)
        beta_val = random_gen(shape=normalized_shape, rand_min=-100, rand_max=100)
        with Function({}) as ssa_func:
            res = mb.layer_norm(x=x_val, axes=axes, epsilon=epsilon, gamma=gamma_val, beta=beta_val)
            ref = TestNormalizationLayerNorm._np_layer_norm(x=x_val, axes=axes, epsilon=epsilon, gamma=gamma_val, beta=beta_val)
            assert is_close(ref, res.val)


class TestNormalizationLocalResponseNorm:
    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend):
        x_val = np.array([[[1.0, -7.0], [5.0, -6.0], [-3.0, -5.0]]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return [
                mb.local_response_norm(x=x, size=2),
                mb.local_response_norm(x=x, size=3, alpha=0.0001, beta=0.75, k=1.0),
            ]

        expected_output_types = [(1, 3, 2, types.fp32), (1, 3, 2, types.fp32)]
        expected_outputs = [
            np.array(
                [
                    [
                        [0.99996257, -6.98716545],
                        [4.99531746, -5.99191284],
                        [-2.99898791, -4.99531746],
                    ]
                ],
                dtype=np.float32,
            ),
            np.array(
                [
                    [
                        [0.99997497, -6.99143696],
                        [4.99687672, -5.99460602],
                        [-2.99932504, -4.99687672],
                    ]
                ],
                dtype=np.float32,
            ),
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.skipif(not testing_reqs._HAS_TORCH, reason="PyTorch not found.")
    @pytest.mark.parametrize(
        "use_cpu_only, backend, rank, size, alpha, beta, k",
        itertools.product(
            [True, False],
            backends,
            [rank for rank in range(3, 6)],
            [2, 3, 5],
            [0.0001, 0.01],
            [0.75, 1.0],
            [1.0, 2.0],
        ),
    )
    def test_builder_to_backend_stress(
        self, use_cpu_only, backend, rank, size, alpha, beta, k
    ):
        shape = np.random.randint(low=2, high=5, size=rank)
        x_val = random_gen(shape=shape)
        input_placeholders = {"x": mb.placeholder(shape=x_val.shape)}
        input_values = {"x": x_val}

        def build(x):
            return mb.local_response_norm(x=x, size=size, alpha=alpha, beta=beta, k=k)

        torch_lrn = torch.nn.LocalResponseNorm(size=size, alpha=alpha, beta=beta, k=k)
        expected_outputs = [torch_lrn(torch.as_tensor(x_val)).numpy()]
        expected_output_types = [o.shape[:] + (types.fp32,) for o in expected_outputs]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
            atol=1e-2,
            rtol=1e-3,
        )



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

import scipy
from scipy import special
from coremltools.converters.mil import testing_reqs
from coremltools.converters.mil.mil import get_new_symbol
from coremltools.converters.mil.testing_reqs import *

from .testing_utils import run_compare_builder

backends = testing_reqs.backends


class TestReduction:
    # All ops in this test share the same backends
    @pytest.mark.parametrize(
        "use_cpu_only, backend, mode",
        itertools.product(
            [True, False],
            backends,
            [
                "argmax",
                "argmin",
                "l1_norm",
                "l2_norm",
                "log_sum",
                "log_sum_exp",
                "max",
                "mean",
                "min",
                "prod",
                "sum",
                "sum_square",
            ],
        ),
    )
    def test_builder_to_backend_smoke(self, use_cpu_only, backend, mode):
        val = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=val.shape)}
        input_values = {"x": val}

        if mode in {"argmax", "argmin"}:
            expected_output_types = (2, types.int32)
        else:
            expected_output_types = (2, types.fp32)

        if mode == "argmax":
            build = lambda x: mb.reduce_argmax(x=x, axis=1, keep_dims=False)
            expected_outputs = np.array([2, 2], dtype=np.int32)
        elif mode == "argmin":
            build = lambda x: mb.reduce_argmin(x=x, axis=1, keep_dims=False)
            expected_outputs = np.array([0, 0], dtype=np.int32)
        elif mode == "l1_norm":
            build = lambda x: mb.reduce_l1_norm(x=x, axes=[1], keep_dims=False)
            expected_outputs = np.array([6.0, 15.0], dtype=np.float32)
        elif mode == "l2_norm":
            build = lambda x: mb.reduce_l2_norm(x=x, axes=[1], keep_dims=False)
            expected_outputs = np.array([3.74165738, 8.77496438], dtype=np.float32)
        elif mode == "log_sum":
            build = lambda x: mb.reduce_log_sum(x=x, axes=[1], keep_dims=False)
            expected_outputs = np.array([1.7917595, 2.70805025], dtype=np.float32)
        elif mode == "log_sum_exp":
            build = lambda x: mb.reduce_log_sum_exp(x=x, axes=[1], keep_dims=False)
            expected_outputs = np.array([3.40760589, 6.40760612], dtype=np.float32)
        elif mode == "max":
            build = lambda x: mb.reduce_max(x=x, axes=[1], keep_dims=False)
            expected_outputs = np.array([3.0, 6.0], dtype=np.float32)
        elif mode == "mean":
            build = lambda x: mb.reduce_mean(x=x, axes=[1], keep_dims=False)
            expected_outputs = np.array([2.0, 5.0], dtype=np.float32)
        elif mode == "min":
            build = lambda x: mb.reduce_min(x=x, axes=[1], keep_dims=False)
            expected_outputs = np.array([1.0, 4.0], dtype=np.float32)
        elif mode == "prod":
            build = lambda x: mb.reduce_prod(x=x, axes=[1], keep_dims=False)
            expected_outputs = np.array([6.0, 120.0], dtype=np.float32)
        elif mode == "sum":
            build = lambda x: mb.reduce_sum(x=x, axes=[1], keep_dims=False)
            expected_outputs = np.array([6.0, 15.0], dtype=np.float32)
        elif mode == "sum_square":
            build = lambda x: mb.reduce_sum_square(x=x, axes=[1], keep_dims=False)
            expected_outputs = np.array([14.0, 77.0], dtype=np.float32)
        else:
            raise NotImplementedError()

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend, mode",
        itertools.product([True, False], backends, ["max", "mean"]),
    )
    def test_builder_to_backend_global_pool_2d(self, use_cpu_only, backend, mode):
        # test lowering to spatial reduction to global_pool path
        val = np.array([[[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=val.shape)}
        input_values = {"x": val}

        expected_output_types = (1, 1, 1, 1, types.fp32)

        if mode == "max":
            build = lambda x: mb.reduce_max(x=x, axes=[2, -1], keep_dims=True)
            expected_outputs = np.array([[[[6.0]]]], dtype=np.float32)
        elif mode == "mean":
            build = lambda x: mb.reduce_mean(x=x, axes=[3, -2], keep_dims=True)
            expected_outputs = np.array([[[[3.5]]]], dtype=np.float32)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.parametrize(
        "use_cpu_only, backend, mode",
        itertools.product([True, False], backends, ["max", "mean"]),
    )
    def test_builder_to_backend_global_pool_3d(self, use_cpu_only, backend, mode):
        # test lowering to spatial reduction to global_pool path
        val = np.array([[[[[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]]]]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=val.shape)}
        input_values = {"x": val}

        expected_output_types = (1, 1, 1, 1, 1, types.fp32)

        if mode == "max":
            build = lambda x: mb.reduce_max(x=x, axes=[2, -1, 3], keep_dims=True)
            expected_outputs = np.array([[[[[6.0]]]]], dtype=np.float32)
        elif mode == "mean":
            build = lambda x: mb.reduce_mean(x=x, axes=[-3, 3, 4], keep_dims=True)
            expected_outputs = np.array([[[[[3.5]]]]], dtype=np.float32)

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            backend=backend,
        )

    @pytest.mark.parametrize(
        ["axis", "keep_dims"], itertools.product([1, -3], [True, False])
    )
    def test_builder_eval(self, axis, keep_dims):
        x_val = random_gen(shape=(1, 3, 4, 4), rand_min=-100.0, rand_max=100.0)

        @ssa_fn
        def test_reduce_argmax():
            res = mb.reduce_argmax(x=x_val, axis=axis, keep_dims=keep_dims).val
            ref = np.argmax(x_val, axis=axis)
            assert is_close(ref, res)

        @ssa_fn
        def test_reduce_argmin():
            res = mb.reduce_argmin(x=x_val, axis=axis, keep_dims=keep_dims).val
            ref = np.argmin(x_val, axis=axis)
            assert is_close(ref, res)

        @ssa_fn
        def test_reduce_l1_norm():
            res = mb.reduce_l1_norm(x=x_val, axes=[axis], keep_dims=keep_dims).val
            ref = np.sum(np.abs(x_val), axis=axis, keepdims=keep_dims)
            assert is_close(ref, res)

        @ssa_fn
        def test_reduce_l2_norm():
            res = mb.reduce_l2_norm(x=x_val, axes=[axis], keep_dims=keep_dims).val
            ref = np.sqrt(np.sum(np.square(x_val), axis=axis, keepdims=keep_dims))
            assert is_close(ref, res)

        @ssa_fn
        def test_reduce_log_sum():
            x_val = random_gen(shape=(1, 3, 4, 4), rand_min=0.0, rand_max=100.0)
            res = mb.reduce_log_sum(x=x_val, axes=[axis], keep_dims=keep_dims).val
            ref = np.log(np.sum(x_val, axis=axis, keepdims=keep_dims))
            assert is_close(ref, res)

        @ssa_fn
        def test_reduce_log_sum_exp():
            res = mb.reduce_log_sum_exp(x=x_val, axes=[axis], keep_dims=keep_dims).val
            ref = scipy.special.logsumexp(x_val, axis=axis, keepdims=keep_dims)
            assert is_close(ref, res)

        @ssa_fn
        def test_reduce_max():
            res = mb.reduce_max(x=x_val, axes=[axis], keep_dims=keep_dims).val
            ref = np.max(x_val, axis=axis, keepdims=keep_dims)
            assert is_close(ref, res)

        @ssa_fn
        def test_reduce_mean():
            res = mb.reduce_mean(x=x_val, axes=[axis], keep_dims=keep_dims).val
            ref = np.mean(x_val, axis=axis, keepdims=keep_dims)
            assert is_close(ref, res)

        @ssa_fn
        def test_reduce_min():
            res = mb.reduce_min(x=x_val, axes=[axis], keep_dims=keep_dims).val
            ref = np.min(x_val, axis=axis, keepdims=keep_dims)
            assert is_close(ref, res)

        @ssa_fn
        def test_reduce_prod():
            res = mb.reduce_prod(x=x_val, axes=[axis], keep_dims=keep_dims).val
            ref = np.prod(x_val, axis=axis, keepdims=keep_dims)
            assert is_close(ref, res)

        @ssa_fn
        def test_reduce_sum():
            res = mb.reduce_sum(x=x_val, axes=[axis], keep_dims=keep_dims).val
            ref = np.sum(x_val, axis=axis, keepdims=keep_dims)
            assert is_close(ref, res)

        @ssa_fn
        def test_reduce_sum_square():
            res = mb.reduce_sum_square(x=x_val, axes=[axis], keep_dims=keep_dims).val
            ref = np.sum(np.square(x_val), axis=axis, keepdims=keep_dims)
            assert is_close(ref, res)

        test_reduce_argmax()
        test_reduce_argmin()
        test_reduce_l1_norm()
        test_reduce_l2_norm()
        test_reduce_log_sum()
        test_reduce_log_sum_exp()
        test_reduce_max()
        test_reduce_mean()
        test_reduce_min()
        test_reduce_prod()
        test_reduce_sum()
        test_reduce_sum_square()

    @pytest.mark.parametrize(
        "use_cpu_only, backend", itertools.product([True, False], backends,)
    )
    def test_builder_to_backend_symbolic(self, use_cpu_only, backend):
        # TODO: variadic (rdar://59559656)

        s0 = get_new_symbol()

        val = np.array([[1.0, 2.0, 3.0], [4.0, 5.0, 6.0]], dtype=np.float32)
        input_placeholders = {"x": mb.placeholder(shape=(s0, 3))}
        input_values = {"x": val}

        def build(x):
            return [
                mb.reduce_argmax(x=x, axis=1, keep_dims=True),
                mb.reduce_argmin(x=x, axis=0, keep_dims=True),
            ]

        expected_output_types = [(s0, 1, types.int32), (1, 3, types.int32)]
        expected_outputs = [
            np.array([[2], [2]], dtype=np.int32),
            np.array([[0], [0], [0]], dtype=np.int32),
        ]

        run_compare_builder(
            build,
            input_placeholders,
            input_values,
            expected_output_types,
            expected_outputs,
            use_cpu_only=use_cpu_only,
            frontend_only=False,
            backend=backend,
        )



# Copyright (c) 2019, Apple Inc. All rights reserved.
#
# Use of this source code is governed by a BSD-3-clause license that can be
# found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from ..._deps import _HAS_SKLEARN
from ...models import MLModel as _MLModel
import coremltools

# from coremltools.proto import Model_pb2
from coremltools.proto import FeatureTypes_pb2

if _HAS_SKLEARN:
    import sklearn.neighbors as _neighbors
    from . import _sklearn_util

import numpy as np
import scipy as sp

model_type = "classifier"
sklearn_class = _neighbors.KNeighborsClassifier


def convert(model, input_name, output_name):
    """Convert a scikit KNeighborsClassifier to protobuf format.

    Parameters
    ----------
    model : KNeighborsClassifier
        A trained scikit-learn KNeighborsClassifier model.

    input_name: str
        Name of the input column.

    output_name: str
        Name of the output column.

    Returns
    -------
    model_spec: An object of type Model_pb.
        Protobuf representation of the model
    """
    if not (_HAS_SKLEARN):
        raise RuntimeError(
            "scikit-learn not found. scikit-learn conversion API is disabled."
        )

    _sklearn_util.check_expected_type(model, sklearn_class)

    _check_fitted(model)
    _check_algorithm(model)
    _check_weighting_scheme(model)
    _check_distance_metric(model)

    return _MLModel(_convert_k_neighbors_classifier(model, input_name, output_name))


def supports_output_scores(model):
    """KNeighborsClassifier models do not support output scores."""
    return False


def get_output_classes(model):
    """Get the candidate classes for the model."""
    _check_fitted(model)
    return list(model.classes_)


def _convert_k_neighbors_classifier(model, input_name, output_name):
    """Convert the scikit KNeighborsClassifier to CoreML. Assumes initial validation of the scikit model has been done."""

    spec = coremltools.proto.Model_pb2.Model()
    spec.specificationVersion = coremltools.SPECIFICATION_VERSION

    spec.kNearestNeighborsClassifier.numberOfNeighbors.defaultValue = model.n_neighbors
    spec.kNearestNeighborsClassifier.numberOfNeighbors.range.minValue = 1
    spec.kNearestNeighborsClassifier.numberOfNeighbors.range.maxValue = _number_of_samples(
        model, spec
    )  # is there a better heuristic to use here?

    number_of_dimensions = 0
    if _is_algorithm_brute(model):
        number_of_dimensions = model._fit_X.shape[1]
        spec.kNearestNeighborsClassifier.nearestNeighborsIndex.linearIndex.MergeFromString(
            b""
        )
    elif _is_algorithm_kd_tree(model):
        npdata = np.asarray(model._tree.data)
        number_of_dimensions = get_input_dimension(model)
        spec.kNearestNeighborsClassifier.nearestNeighborsIndex.singleKdTreeIndex.leafSize = (
            model.leaf_size
        )
    else:
        raise TypeError(
            "KNeighbors algorithm not supported for CoreML conversion: {}".format(
                model.algorithm
            )
        )
    spec.kNearestNeighborsClassifier.nearestNeighborsIndex.numberOfDimensions = (
        number_of_dimensions
    )

    # Make sure the distance function is set
    spec.kNearestNeighborsClassifier.nearestNeighborsIndex.squaredEuclideanDistance.MergeFromString(
        b""
    )

    input_features = spec.description.input.add()
    input_features.name = input_name[0][0]
    input_features.type.multiArrayType.shape.extend([number_of_dimensions])
    input_features.type.multiArrayType.dataType = (
        FeatureTypes_pb2.ArrayFeatureType.FLOAT32
    )

    output_label = spec.description.output.add()
    output_label.name = output_name[0][0]

    # predictedFeatureName is required since KNN is a classifier and it should be same as outputName.
    spec.description.predictedFeatureName = output_label.name

    # Need to confirm if scikit only accepts integer labels
    output_label.type.int64Type.MergeFromString(b"")
    spec.kNearestNeighborsClassifier.uniformWeighting.MergeFromString(b"")

    _extract_training_data(model, spec)

    return spec


def _number_of_samples(model, spec):
    """Get the number of samples the model is fitted to."""

    if _is_algorithm_brute(model):
        return model._fit_X.shape[0]
    elif _is_algorithm_kd_tree(model):
        return len(np.asarray(model._tree.data))
    return 0


def _extract_training_data(model, spec):
    """Extract the training data from the scikit model and add it to the CoreML spec"""

    if _is_algorithm_brute(model):
        X = model._fit_X
        if _is_valid_sparse_format(X):
            X = _unpack_sparse(X)

        for sample in X:
            coreml_sample = (
                spec.kNearestNeighborsClassifier.nearestNeighborsIndex.floatSamples.add()
            )
            for feature in sample:
                coreml_sample.vector.append(feature)

    elif _is_algorithm_kd_tree(model):
        # sklearn guarantees that tree data is not stored in a sparse format
        npdata = np.asarray(model._tree.data)
        for sample in npdata:
            coreml_sample = (
                spec.kNearestNeighborsClassifier.nearestNeighborsIndex.floatSamples.add()
            )
            for feature in sample:
                coreml_sample.vector.append(feature)

    for label in model._y:
        spec.kNearestNeighborsClassifier.int64ClassLabels.vector.append(label)


def get_input_dimension(model):
    """Get the input dimension for the model"""
    _check_fitted(model)
    number_of_dimensions = 0
    if _is_algorithm_brute(model):
        number_of_dimensions = model._fit_X.shape[1]
    elif _is_algorithm_kd_tree(model):
        npdata = np.asarray(model._tree.data)
        number_of_dimensions = len(npdata[0])
    else:
        raise TypeError(
            "KNeighbors algorithm not supported for CoreML conversion: {}".format(
                model.algorithm
            )
        )
    return number_of_dimensions


def _check_fitted(model):
    """Simple wrapper to check if the KNeighborsClassifier has been fitted."""
    return _sklearn_util.check_fitted(
        model, lambda m: hasattr(m, "_fit_method") or hasattr(m, "_fit_X")
    )


def _check_algorithm(model):
    """Ensure the kNeighbors algorithm for the given scikit model is a supported type"""
    is_valid = False
    print_name = ""
    if model.algorithm == "brute" or model.algorithm == "kd_tree":
        is_valid = True
        print_name = model.algorithm
    elif model.algorithm == "auto" and model._fit_method == "kd_tree":
        is_valid = True
        print_name = "kd_tree"
    elif model.algorithm == "auto" and model._fit_method == "brute":
        is_valid = True
        print_name = "brute"
    if not is_valid:
        raise TypeError(
            "KNeighbors algorithm not supported for CoreML conversion: {}".format(
                print_name
            )
        )


def _check_weighting_scheme(model):
    """Simple wrapper to ensure the weighting scheme is valid for CoreML conversion"""
    is_valid = False
    if model.weights == "uniform":
        is_valid = True

    # Other cases CoreML doesn't support include weighting by distance or a user-provided 'callable' object.

    if not is_valid:
        print_name = ""
        if _is_printable(model.weights):
            print_name = model.weights
        else:
            print_name = getattr(model.weights, "__name__", repr(model.weights))
        raise TypeError(
            "KNeighbors weight function not supported for CoreML conversion: {}".format(
                print_name
            )
        )


def _check_distance_metric(model):
    """Simple wrapper to ensure the distance metric is valid for CoreML conversion"""
    is_valid = False
    if model.metric == "euclidean":
        is_valid = True
    elif model.metric == "minkowski" and model.p == 2:
        is_valid = True

    # There are a number of other distance metrics supported by scikit that CoreML doesn't currently support.

    if not is_valid:
        print_name = ""
        if _is_printable(model.metric):
            print_name = model.metric
        else:
            print_name = getattr(model.metric, "__name__", repr(model.metric))
        raise TypeError(
            "KNeighbors distance metric not supported for CoreML conversion: {}".format(
                print_name
            )
        )


def _is_algorithm_brute(model):
    """Checks if the algorithm for the scikit model is set to 'brute'."""
    return model.algorithm == "brute" or (
        model.algorithm == "auto" and model._fit_method == "brute"
    )


def _is_algorithm_kd_tree(model):
    """Checks if the algorithm for the scikit model is set to 'kd_tree'."""
    return model.algorithm == "kd_tree" or (
        model.algorithm == "auto" and model._fit_method == "kd_tree"
    )


def _is_printable(obj):
    """Check if the object is a valid text type."""
    return isinstance(obj, str)


def _is_valid_sparse_format(obj):
    """Check if the object is in CSR sparse format (the only valid type for KNeighborsClassifier)"""
    return isinstance(obj, sp.sparse.csr_matrix)


def _unpack_sparse(obj):
    """Unpack the sparse matrix into a format that we can easily iterate over for insertion into a CoreML model."""
    if not sp.sparse.issparse(obj):
        raise TypeError("Object {} is not a scipy sparse matrix type".format(type(obj)))
    return obj.toarray()



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

_TF_OPS_REGISTRY = {}


def register_tf_op(_func=None, tf_alias=None, override=False):
    """
    Registration routine for TensorFlow operators
    _func: (TF conversion function) [Default=None]
        TF conversion function to register

    tf_alias: (List of string) [Default=None]
        All other TF operators that should also be mapped to
        current conversion routine.
        e.g. Sort aliased with SortV1, SortV2
        All provided alias operators must not be registered previously.

    override: (Boolean) [Default=False]
        If True, overrides earlier registration i.e. specified
        operator and alias will start pointing to current conversion
        function.
        Otherwise, duplicate registration will error out.
    """

    def func_wrapper(func):
        f_name = func.__name__

        if not override and f_name in _TF_OPS_REGISTRY:
            raise ValueError("TF op {} already registered.".format(f_name))
        _TF_OPS_REGISTRY[f_name] = func
        # If tf_alias is provided, then all the functions mentioned as aliased
        # are mapped to current function
        if tf_alias is not None:
            for name in tf_alias:
                if not override and name in _TF_OPS_REGISTRY:
                    msg = "TF op alias {} already registered."
                    raise ValueError(msg.format(name))
                _TF_OPS_REGISTRY[name] = func
        return func

    if _func is None:
        # decorator called without argument
        return func_wrapper
    return func_wrapper(_func)



from __future__ import absolute_import as _
from __future__ import division as _
from __future__ import print_function as _
from __future__ import unicode_literals as _

import numpy as np
import copy

from typing import Sequence, Callable, List, Tuple, Optional, Text, Any
from coremltools.models.neural_network import NeuralNetworkBuilder  # type: ignore
from onnx import TensorProto
from ._graph import Node, Graph
from coremltools.proto import NeuralNetwork_pb2  # type: ignore
from ._error_utils import ErrorHandling

from ._operators import (
    _convert_abs,
    _convert_relu,
    _convert_sqrt,
    _convert_exp,
    _convert_elu,
    _convert_selu,
    _convert_sigmoid,
    _convert_sign,
    _convert_prelu,
    _convert_upsample,
    _convert_softsign,
    _convert_softplus,
    _convert_log,
    _convert_neg,
    _convert_reciprocal,
    _convert_hardsigmoid,
    _convert_reorganize_data,
    _add_pool,
    _get_pool_params,
    _add_conv,
    _get_conv_params,
    _convert_thresholdedrelu,
    _convert_leaky_relu,
    _convert_lrn,
)

from ._operators import _convert_pad as _convert_pad_5d

INT_MAX = 2 ** 63 - 1


## Helper functions
def load_input_constants(builder, node, graph, err):
    for i in range(len(node.inputs)):
        if (
            node.inputs[i] in node.input_tensors
            and node.inputs[i] not in graph.constants_loaded
        ):
            value = node.input_tensors[node.inputs[i]]
            builder.add_load_constant_nd(
                name=node.name + "_load_constant_" + str(i),
                output_name=node.inputs[i],
                constant_value=value,
                shape=[1] if value.shape == () else value.shape,
            )
            graph.constants_loaded.add(node.inputs[i])


def _add_conv_like_op(
    add_func, get_params_func, params_dict, builder, node, graph, err
):
    rank = builder._get_rank(node.inputs[0])
    if rank == 4:
        get_params_func(builder, node, graph, err, params_dict)
        add_func(
            node.inputs,
            node.outputs,
            params_dict=params_dict,
            builder=builder,
            node=node,
            graph=graph,
            err=err,
        )
    elif rank == 3:
        axes = [0, 3]
        # Make 5d tensor
        expanded_node_output = node.name + "_" + node.inputs[0] + "_expanded"
        builder.add_expand_dims(
            name=node.name + "_ip_expand",
            input_name=node.inputs[0],
            output_name=expanded_node_output,
            axes=axes,
        )
        node.inputs[0] = expanded_node_output
        output_name = node.outputs[0]
        node.outputs[0] = node.name + "_" + output_name + "_expanded"
        # Add conversion op
        get_params_func(builder, node, graph, err, params_dict, axis="width")
        add_func(
            node.inputs,
            node.outputs,
            params_dict=params_dict,
            builder=builder,
            node=node,
            graph=graph,
            err=err,
        )
        # Make 3d tensor back
        builder.add_squeeze(
            name=node.name + "_ip_squeeze_out",
            input_name=node.outputs[0],
            output_name=output_name,
            axes=axes,
        )
    else:
        return err.unsupported_op_configuration(
            builder, node, graph, "provided number axes {} not supported".format(rank)
        )


def add_broadcastable_op_chain(builder, node, err, add_op_function):
    """
    Splits list of input into chain of operator with two inputs
    where output of first node is fed into next one until the final input
    is processed
    Pass node:            Node to be converted
         add_op_function: Conversion function to be used
    """
    total_nodes = len(node.inputs)

    if total_nodes < 2:
        # TODO: Skip or CopyProp + DeadCode elimination
        builder.add_activation(
            name=node.name,
            non_linearity="LINEAR",
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            params=[1.0, 0.0],
        )
    elif total_nodes == 2:
        add_op_function(
            name=node.name, input_names=node.inputs, output_name=node.outputs[0]
        )
    else:
        decorator = 0
        out_name = node.outputs[0]
        # Add broadcastable layer for first two inputs
        add_op_function(
            name=node.name,
            input_names=[node.inputs[0], node.inputs[1]],
            output_name=out_name + "_" + str(decorator),
        )
        # Continue chain of broadcastable layers
        for i in range(2, total_nodes - 1):
            add_op_function(
                name=node.name,
                input_names=[out_name + "_" + str(decorator), node.inputs[i]],
                output_name=out_name + "_" + str(decorator + 1),
            )
            decorator += 1
        # End chain of broadcastable layers with final output
        add_op_function(
            name=node.name + "_" + str(decorator),
            input_names=[out_name + "_" + str(decorator), node.inputs[total_nodes - 1]],
            output_name=out_name,
        )


def add_bn_with_expansion(
    builder,
    node,
    err,
    node_name,
    input_name,
    output_name,
    channels,
    scale,
    bias,
    mean=None,
    var=None,
    epsilon=None,
    compute_mean_var=False,
    instance_normalization=False,
    axes_for_expansion=[],
):
    real_input_name = input_name
    real_output_name = output_name

    # Expand input if needed
    if len(axes_for_expansion) != 0:
        input_name = node_name + "_" + input_name + "_expanded"
        output_name = output_name + "_expanded"
        builder.add_expand_dims(
            name=node_name + "_expand",
            input_name=real_input_name,
            output_name=input_name,
            axes=axes_for_expansion,
        )

    builder.add_batchnorm(
        name=node.name,
        channels=channels,
        gamma=scale,
        beta=bias,
        mean=mean,
        variance=var,
        input_name=input_name,
        output_name=output_name,
        compute_mean_var=compute_mean_var,
        instance_normalization=instance_normalization,
        epsilon=epsilon,
    )

    # Squeeze output if needed
    if len(axes_for_expansion) != 0:
        builder.add_squeeze(
            name=node_name + "_squeeze",
            input_name=output_name,
            output_name=real_output_name,
            axes=axes_for_expansion,
        )


# Helper function to convert RandomNormal, RandomUniform and it's variants
def add_random(builder, node, graph, err, add_op_function):
    # Ignoring attribute `dtype` as CoreML internally represents tensors into 'Float'
    mean = node.attrs.get("mean", 0.0)
    scale = node.attrs.get("scale", 1.0)
    seed = node.attrs.get("seed", -1)
    shape = node.attrs.get("shape", None)
    if shape is None:
        return err.unsupported_op_configuration(
            builder, node, graph, "Shape not provided"
        )
    add_op_function(
        name=node.name,
        output_name=node.outputs[0],
        output_shape=shape,
        mean=mean,
        stddev=scale,
        seed=seed,
    )


## Converter functions


def _convert_acos(builder, node, graph, err):
    """
    convert to CoreML Acos Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3793
    """
    load_input_constants(builder, node, graph, err)
    builder.add_acos(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_acosh(builder, node, graph, err):
    """
    convert to CoreML Acosh Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3925
    """
    load_input_constants(builder, node, graph, err)
    builder.add_acosh(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_add(builder, node, graph, err):
    """
    convert to CoreML Add Broadcastable Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4117
    """
    load_input_constants(builder, node, graph, err)
    add_broadcastable_op_chain(builder, node, err, builder.add_add_broadcastable)


def _convert_argmax(builder, node, graph, err):
    """
    convert to CoreML ArgMax Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4961
    """
    axis = node.attrs.get("axis", 0)
    keepdims = node.attrs.get("keepdims", True)
    builder.add_argmax(
        name=node.name,
        input_name=node.inputs[0],
        output_name=node.outputs[0],
        axis=axis,
        keepdims=keepdims,
    )


def _convert_argmin(builder, node, graph, err):
    """
    convert to CoreML ArgMin Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4988
    """
    axis = node.attrs.get("axis", 0)
    keepdims = node.attrs.get("keepdims", True)
    builder.add_argmin(
        name=node.name,
        input_name=node.inputs[0],
        output_name=node.outputs[0],
        axis=axis,
        keepdims=keepdims,
    )


def _convert_asin(builder, node, graph, err):
    """
    convert to CoreML Asin Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3771
    """
    load_input_constants(builder, node, graph, err)
    builder.add_asin(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_asinh(builder, node, graph, err):
    """
    convert to CoreML Asinh Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3903
    """
    load_input_constants(builder, node, graph, err)
    builder.add_asinh(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_atan(builder, node, graph, err):
    """
    convert to CoreML Atan Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3815
    """
    load_input_constants(builder, node, graph, err)
    builder.add_atan(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_atanh(builder, node, graph, err):
    """
    convert to CoreML Atanh Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3947
    """
    load_input_constants(builder, node, graph, err)
    builder.add_atanh(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_bn(builder, node, graph, err):
    """
    convert to CoreML BatchNorm Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L1633
    """
    if len(node.outputs) > 1:
        return err.unsupported_op_configuration(
            builder,
            node,
            graph,
            "This converter only supports BatchNormalization with one output",
        )

    epsilon = node.attrs.get("epsilon", 1e-5)
    scale_name = node.inputs[1]

    if scale_name in node.input_tensors:
        channels = node.input_tensors[scale_name].shape
    elif scale_name in graph.shape_dict:
        channels = graph.shape_dict[scale_name]
    else:
        err.unsupported_op_configuration(
            builder, node, graph, "Input shape not available"
        )

    # TODO: Move error check under VERBOSE / DEBUG Mode
    for i in range(2, len(node.inputs)):
        ip_name = node.inputs[i]
        if ip_name in node.input_tensors:
            tensor_shape = node.input_tensors[ip_name].shape
        else:
            if ip_name not in graph.shape_dict:
                return err.unsupported_op_configuration(
                    builder, node, graph, "Input shape not available"
                )
            tensor_shape = graph.shape_dict[ip_name]
        if tensor_shape != channels:
            err.unsupported_op_configuration(
                builder,
                node,
                graph,
                "Shape mismatch between Scale, Bias, Mean and Variance",
            )

    scale = (
        node.input_tensors[node.inputs[1]]
        if node.inputs[1] in node.input_tensors
        else np.ones(shape=channels, dtype=np.float32)
    )
    bias = (
        node.input_tensors[node.inputs[2]]
        if node.inputs[2] in node.input_tensors
        else np.zeros(shape=channels, dtype=np.float32)
    )
    mean = (
        node.input_tensors[node.inputs[3]]
        if node.inputs[3] in node.input_tensors
        else np.zeros(shape=channels, dtype=np.float32)
    )
    var = (
        node.input_tensors[node.inputs[4]]
        if node.inputs[4] in node.input_tensors
        else np.ones(shape=channels, dtype=np.float32)
    )

    rank = builder._get_rank(node.inputs[0])
    # ONNX converts B x C tensor into B x C x 1 hence
    # Rank 2 BN is mapped to Rank 3 BN
    if rank == 3:
        # 1D Batch Norm
        add_bn_with_expansion(
            builder,
            node,
            err,
            node.name,
            node.inputs[0],
            node.outputs[0],
            channels[0],
            scale,
            bias,
            mean,
            var,
            epsilon,
            axes_for_expansion=[0, 3],
        )
    elif rank == 4:
        # 2D Batch Norm
        add_bn_with_expansion(
            builder,
            node,
            err,
            node.name,
            node.inputs[0],
            node.outputs[0],
            channels[0],
            scale,
            bias,
            mean,
            var,
            epsilon,
            axes_for_expansion=[],
        )
    else:
        # Unsupported 1D, 3D and above
        err.unsupported_op_configuration(
            builder, node, graph, "provided number axes {} not supported".format(rank)
        )


def _convert_cast(builder, node, graph, err):
    """
    Perform cast operation in CoreML
        e.g. Casting from Float (assumed) to Int maps to Floor Layer
             For Others, add copy layer
    """
    convert_to = node.attrs.get("to")
    convert_to_int = set(
        {
            TensorProto.UINT8,
            TensorProto.INT8,
            TensorProto.UINT16,
            TensorProto.INT32,
            TensorProto.INT64,
            TensorProto.UINT32,
            TensorProto.UINT64,
        }
    )

    ## TODO: Add support for conversion from STRING TO FLOAT
    ## Currently, such input will error out in parsing
    if convert_to in convert_to_int:
        builder.add_floor(
            name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
        )
    else:
        load_input_constants(builder, node, graph, err)
        builder.add_activation(
            name=node.name,
            non_linearity="LINEAR",
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            params=[1.0, 0.0],
        )


def _convert_ceil(builder, node, graph, err):
    """
    convert to CoreML Ceil Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L5018
    """
    builder.add_ceil(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0],
    )


def _convert_clip(builder, node, graph, err):
    """
    convert to CoreML Clip Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L5066
    """
    max_value = node.attrs.get("max", 3.4028234663852886e38)
    min_value = node.attrs.get("min", -3.4028234663852886e38)
    builder.add_clip(
        name=node.name,
        input_name=node.inputs[0],
        output_name=node.outputs[0],
        min_value=min_value,
        max_value=max_value,
    )


def _convert_concat(builder, node, graph, err):
    """
    convert to CoreML ConcatND Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3521
    """
    axis = node.attrs.get("axis")
    load_input_constants(builder, node, graph, err)

    # TODO: Adding Linear layer will change to
    #       either: Skip the op right away
    #       or:     Insert Linear and perform copy-propogation followed by dead code elimination
    if len(node.inputs) == 1:
        builder.add_activation(
            name=node.name,
            non_linearity="LINEAR",
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            params=[1.0, 0.0],
        )
    else:
        builder.add_concat_nd(
            name=node.name,
            input_names=node.inputs,
            output_name=node.outputs[0],
            axis=axis,
        )


def _convert_constant(builder, node, graph, err):
    """
    convert to CoreML Load Constant ND Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3596
    """
    value = node.attrs["value"]
    # HACK: If Value is 0-Rank then make it 1-Rank
    builder.add_load_constant_nd(
        name=node.name,
        output_name=node.outputs[0],
        constant_value=value,
        shape=[1] if value.shape == () else value.shape,
    )
    graph.constants_loaded(node.outputs[0])


def _convert_constant_of_shape(builder, node, graph, err):
    """
    convert to CoreML Fill Static Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3641
    """
    value = node.attrs.get("value", [0.0])
    # if shape is known, create tensor of given shape
    # otherwise create tensor at runtime
    if node.inputs[0] in node.input_tensors:
        output_shape = node.input_tensors[node.inputs[0]]
        # add_fill_static requires shape to be more than rank-1
        if len(output_shape.shape) == 1:
            output_shape = output_shape.reshape(output_shape.shape[0], 1)
        builder.add_fill_static(
            name=node.name,
            output_name=node.outputs[0],
            output_shape=output_shape,
            value=value[0],
        )
    else:
        builder.add_fill_dynamic(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            value=value[0],
        )


def _convert_conv(builder, node, graph, err):
    """
    convert to CoreML Convolution Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L1418
    """
    params_dict = dict()
    params_dict["is_deconv"] = False
    if node.op_type.endswith("Transpose"):
        params_dict["is_deconv"] = True
    # get weights for convolution
    weight_name = node.inputs[1]
    W = None
    if weight_name in node.input_tensors:
        W = node.input_tensors[weight_name]
        params_dict["w_shape"] = W.shape
    else:
        # W is provided as a input
        # Make W compatible for CoreML Conv Layer
        # W ONNX format: OC x KC x H x W
        # Expected CoreML Format: H x W x KC x OC
        W_name = node.inputs[1]
        W_shape = graph.shape_dict[W_name]
        W_rank = len(W_shape)

        params_dict["w_shape"] = W_shape
        if W_rank == 3:
            expanded_node_name = node.name + "_" + W_name + "_expanded"
            builder.add_expand_dims(
                name=node.name + "_w_expand",
                input_name=W_name,
                output_name=expanded_node_name,
                axes=[-2],
            )
            W_name = expanded_node_name

        # Now Permute the W tensor
        W_transpose_axes = [2, 3, 1, 0]
        # If ConvTranpose then, Kernel and Output channels are shuffled
        if params_dict["is_deconv"]:
            W_transpose_axes = [2, 3, 0, 1]

        builder.add_transpose(
            name=node.name + "_w_transpose",
            axes=W_transpose_axes,
            input_name=W_name,
            output_name=W_name + "_transposed",
        )
        W_name = W_name + "_transposed"
        node.inputs[1] = W_name

    params_dict["W"] = W
    bias = None
    if len(node.inputs) > 2:
        bias = node.input_tensors[node.inputs[2]]
    params_dict["bias"] = bias
    params_dict["groups"] = node.attrs.get("group", 1)

    _add_conv_like_op(
        _add_conv, _get_conv_params, params_dict, builder, node, graph, err
    )


def _convert_cos(builder, node, graph, err):
    """
    convert to CoreML Cos Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3727
    """
    load_input_constants(builder, node, graph, err)
    builder.add_cos(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_cosh(builder, node, graph, err):
    """
    convert to CoreML Cosh Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3859
    """
    load_input_constants(builder, node, graph, err)
    builder.add_cosh(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_div(builder, node, graph, err):
    """
    convert to CoreML Divide Broadcastable Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4180
    """
    load_input_constants(builder, node, graph, err)
    add_broadcastable_op_chain(builder, node, err, builder.add_divide_broadcastable)


def _convert_equal(builder, node, graph, err):
    """
    convert to CoreML Equal Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L961
    """
    load_input_constants(builder, node, graph, err)
    builder.add_equal(
        name=node.name, input_names=node.inputs, output_name=node.outputs[0]
    )


def _convert_erf(builder, node, graph, err):
    """
    convert to CoreML Erf Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L5140
    """
    load_input_constants(builder, node, graph, err)
    builder.add_erf(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_expand(builder, node, graph, err):
    """
    convert to CoreML Broadcast To Static/Dynamic Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4086
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4108
    """
    load_input_constants(builder, node, graph, err)
    if node.inputs[1] in node.input_tensors:
        output_shape = node.input_tensors[node.inputs[1]].astype(np.int64)
        builder.add_broadcast_to_static(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            output_shape=output_shape,
        )
    else:
        builder.add_broadcast_to_dynamic(
            name=node.name, input_names=node.inputs, output_name=node.outputs[0],
        )


def _convert_flatten(builder, node, graph, err):
    """
    convert to CoreML Flatten Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4826
    """
    axis = node.attrs.get("axis", 1)
    builder.add_flatten_to_2d(
        name=node.name,
        input_name=node.inputs[0],
        output_name=node.outputs[0],
        axis=axis,
    )


def _convert_floor(builder, node, graph, err):
    """
    convert to CoreML Floor Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L5040
    """
    builder.add_floor(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_gather(builder, node, graph, err):
    """
    convert to CoreML Gather Along Axis Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4296
    """
    axis = node.attrs.get("axis", 0)

    if len(node.inputs) != 2:
        err.unsupported_op_configuration(
            builder, node, graph, "Error in ONNX model: Gather expects two inputs"
        )

    if (
        node.inputs[0] in node.input_tensors
        and node.inputs[0] not in graph.constants_loaded
    ):
        value = node.input_tensors[node.inputs[0]]
        builder.add_load_constant_nd(
            name=node.name + "_load_data",
            output_name=node.inputs[0],
            constant_value=value,
            shape=[1] if value.shape == () else value.shape,
        )
        graph.constants_loaded.add(node.inputs[0])

    if (
        node.inputs[1] in node.input_tensors
        and node.inputs[1] not in graph.constants_loaded
    ):
        value = node.input_tensors[node.inputs[1]]
        builder.add_load_constant_nd(
            name=node.name + "_load_indices",
            output_name=node.inputs[1],
            constant_value=value,
            shape=[1] if value.shape == () else value.shape,
        )
        graph.constants_loaded.add(node.inputs[1])

    builder.add_gather(
        name=node.name,
        input_names=[node.inputs[0], node.inputs[1]],
        output_name=node.outputs[0],
        axis=axis,
    )


def _convert_gemm(builder, node, graph, err):
    """
    convert to CoreML Tranpose (Optional) and Inner Product Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4180
    """
    # Read attributes
    alpha = node.attrs.get("alpha", 1.0)
    beta = node.attrs.get("beta", 1.0)
    transA = node.attrs.get("transA", False)
    transB = node.attrs.get("transB", False)

    A = node.inputs[0]
    if A in node.input_tensors:
        A_tensor = node.input_tensors[A]
        builder.add_load_constant_nd(
            name=node.name + A + "_const",
            output_name="const_" + A,
            constant_value=A_tensor,
            shape=A_tensor.shape,
        )
        A = "const_" + A

    if alpha != 1.0:
        builder.add_load_constant_nd(
            name=node.name + "_load_alpha",
            output_name="alpha_for_" + A,
            constant_value=np.array([alpha]),
            shape=[1],
        )
        builder.add_multiply_broadcastable(
            name=node.name + "_alphaA",
            input_names=[A, "alpha_for_" + A],
            output_name=A + "_alphaA",
        )
        A = A + "_alphaA"

    B = node.inputs[1]
    C = node.inputs[2]
    if B in node.input_tensors and C in node.input_tensors:
        B = node.input_tensors[B]
        C = node.input_tensors[C]

        if transB:
            B = B.transpose()

        C = C.flatten()
        builder.add_batched_mat_mul(
            name=node.name,
            input_names=[A],
            output_name=node.outputs[0],
            transpose_a=transA,
            weight_matrix_rows=B.shape[0],
            weight_matrix_columns=B.shape[1],
            W=B,
            bias=C,
        )
    else:
        ## TODO: Test coverage when B and C are non-constant
        ## Should C be of Rank-1? or it's okay to keep it that way?
        if beta != 1.0:
            builder.add_load_constant_nd(
                name=node.name + "_load_beta",
                output_name="beta_for_" + B,
                constant_value=np.array([beta]),
                shape=[1],
            )
            builder.add_multiply_broadcastable(
                name=node.name + "_betaC",
                input_names=[C, "beta_for_" + B],
                output_name=C + "_betaC",
            )
            C = C + "_betaC"

        builder.add_batched_mat_mul(
            name=node.name,
            input_names=[A, B],
            output_name=node.outputs[0] + "_b_mat_mul",
            transpose_a=transA,
            transpose_b=transB,
        )

        builder.add_add_broadcastable(
            name=node.name + "_add_bias",
            input_names=[node.outputs[0] + "_b_mat_mul", C],
            output_name=node.outputs[0],
        )


def _convert_greater(builder, node, graph, err):
    """
    convert to CoreML Greater than Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L853
    """
    load_input_constants(builder, node, graph, err)
    builder.add_greater_than(
        name=node.name, input_names=node.inputs, output_name=node.outputs[0],
    )


def _convert_gru(
    builder, node, graph, err
):  # type: (NeuralNetworkBuilder, Node, Graph, ErrorHandling) -> None
    """
    convert to CoreML GRU Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3104
    """

    def get_weights(W, W_name, R, R_name, B):
        """
        Helper routine to return weights in CoreML LSTM required format
        """
        W = np.expand_dims(np.expand_dims(W, 3), 3)
        R = np.expand_dims(np.expand_dims(R, 3), 3)

        if W is None:
            err.missing_initializer(
                node,
                "Weight tensor: {} not found in the graph initializer".format(W_name),
            )
        if R is None:
            err.missing_initializer(
                node,
                "Weight tensor: {} not found in the graph initializer".format(R_name),
            )

        W_z, W_r, W_h = np.split(np.squeeze(W), 3)  # type: ignore
        R_z, R_r, R_h = np.split(np.squeeze(R), 3)  # type: ignore

        W_x = [W_z, W_r, W_h]
        W_h = [R_z, R_r, R_h]
        b = None
        if B is not None:
            b_Wz, b_Wr, b_Wh, b_Rz, b_Rr, b_Rh = np.split(np.squeeze(B), 6)  # type: ignore
            b = [b_Wz + b_Rz, b_Wr + b_Rr, b_Wh + b_Rh]

        return W_x, W_h, b

    def expand_dim(node_name, input_name, output_name, axes):
        builder.add_expand_dims(
            name=node_name, input_name=input_name, output_name=output_name, axes=axes
        )

    # Read attributes
    # activation alpha and beta
    if "activation_alpha" in node.attrs or "activation_beta" in node.attrs:
        err.unsupported_feature_warning(
            node, "Activation parameter alpha and beta are currently not used"
        )

    inner_activation = "SIGMOID"
    output_activation = "TANH"

    if "activations" in node.attrs:
        activations_list = node.attrs["activations"]

        if len(activations_list) < 2:
            err.unsupported_op_configuration(
                builder,
                node,
                graph,
                "Error in ONNX model: Less number of activations provided",
            )

        inner_activation = activations_list[0].upper()
        output_activation = activations_list[1].upper()

    # Extract direction from ONNX attribute
    direction = node.attrs.get("direction", "forward")
    if direction == "bidirectional":
        return err.unsupported_op_configuration(
            builder,
            node,
            graph,
            "Bidirectional GRU not supported!! Please consider adding custom conversion function/layer",
        )

    hidden_size = node.attrs.get("hidden_size")

    # Read inputs
    W_name = node.inputs[1]
    R_name = node.inputs[2]
    B = None
    if len(node.inputs) > 3:
        B_name = node.inputs[3]
        B = node.input_tensors.get(B_name, None)

    if W_name not in node.input_tensors or R_name not in node.input_tensors:
        return err.unsupported_op_configuration(
            builder,
            node,
            graph,
            "Input and Recursion weights must be known!! Please consider adding custom conversion function/layer",
        )

    W = node.input_tensors.get(W_name, None)
    R = node.input_tensors.get(R_name, None)

    # Get weights for forward direction
    W_x, W_h, b = get_weights(W, W_name, R, R_name, B)

    # shape of input
    input_size = W_x[0].shape[1]

    # Get input and output for hidden and cell
    input_h = node.inputs[5] if len(node.inputs) > 5 else node.inputs[0] + "_h_input"
    output_h = (
        node.outputs[1] if len(node.outputs) > 1 else node.outputs[0] + "_h_output"
    )
    output_h_5d = output_h + "_5d"

    if len(node.inputs) < 6:
        # if input is not present in the network, load they as constant
        if node.inputs[0] not in graph.shape_dict:
            err.unsupported_op_configuration(
                builder, node, graph, "Input shape not represented within Graph"
            )

        # Input is represented as [Seq Len, Batch Size, Input Size]
        batch_size = graph.shape_dict[node.inputs[0]][1]
        builder.add_load_constant_nd(
            name=node.name + "_load_initial_h",
            output_name=input_h,
            constant_value=0.0,
            shape=[1, batch_size, hidden_size],
        )

    # CoreML GRU expects 5-d tensor
    # Expand dimensions of input to 5-d for compatibility
    input_rank = builder._get_rank(node.inputs[0])
    if input_rank == -1:
        return err.unsupported_op_configuration(
            builder, node, graph, "Rank unknown for input"
        )

    if input_rank < 5:
        add_nodes = 5 - input_rank

        # TODO: Add one expand instead of adding one after another for input, h
        expand_dim(
            node.name + "_expand_in_0",
            node.inputs[0],
            node.inputs[0] + "_expand_out_0",
            [input_rank],
        )
        expand_dim(
            node.name + "_expand_in_h_0",
            input_h,
            input_h + "_expand_out_h_0",
            [input_rank],
        )

        for i in range(1, add_nodes):
            i_str = str(i)
            i_p_str = str(i - 1)
            expand_dim(
                node.name + "_expand_in_" + i_str,
                node.inputs[0] + "_expand_out_" + i_p_str,
                node.inputs[0] + "_expand_out_" + i_str,
                [input_rank + i],
            )
            expand_dim(
                node.name + "_expand_in_h_" + i_str,
                input_h + "_expand_out_h_" + i_p_str,
                input_h + "_expand_out_h_" + i_str,
                [input_rank + i],
            )

    builder.add_gru(
        name=node.name,
        W_h=W_h,
        W_x=W_x,
        b=b,
        hidden_size=hidden_size,
        input_size=input_size,
        input_names=[
            node.inputs[0] + "_expand_out_" + str(add_nodes - 1),
            input_h + "_expand_out_h_" + str(add_nodes - 1),
        ],
        output_names=[node.outputs[0] + "_5d_out", output_h_5d],
        inner_activation=inner_activation,
        activation=output_activation,
        output_all=True,
        reverse_input=(direction == "reverse"),
    )

    # CoreML output is [Seq Len, Batch Size, Num Dir * Hidden Size, 1, 1]
    # Return output as [Seq Len, Num Dir, Batch Size, Hidden Size]
    # Following steps:
    #       a. Reshape and split hidden size for direction [Seq Len, Batch Size, Num Dir, Hidden Size, 1]
    #       b. Squeeze last dimension [Seq Len, Batch Size, Num Dir, Hidden Size]
    #       c. Permute to fix the order [Seq Len, Num Dir, Batch Size, Hidden Size, 1]
    builder.add_rank_preserving_reshape(
        name=node.name + "_reshape_",
        input_name=node.outputs[0] + "_5d_out",
        output_name=node.outputs[0] + "_5d_reshaped",
        output_shape=[0, 0, 1, -1, 0],
    )

    builder.add_squeeze(
        name=node.name + "_squeeze_out",
        input_name=node.outputs[0] + "_5d_reshaped",
        output_name=node.outputs[0] + "_4d",
        axes=[-1],
    )

    builder.add_transpose(
        name=node.name + "_transpose",
        axes=[0, 2, 1, 3],
        input_name=node.outputs[0] + "_4d",
        output_name=node.outputs[0],
    )

    # Squeeze dimensions of output_h
    builder.add_squeeze(
        name=node.name + "_squeeze_out_h",
        input_name=output_h_5d,
        output_name=output_h,
        axes=[-1, -2],
    )


def _convert_identity(builder, node, graph, err):
    """
    convert to CoreML Linear Activation Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L417
    """
    # TODO: Skip or CopyProp + DeadCode elimination
    builder.add_activation(
        name=node.name,
        non_linearity="LINEAR",
        input_name=node.inputs[0],
        output_name=node.outputs[0],
        params=[1.0, 0.0],
    )


def _convert_instancenorm(
    builder, node, graph, err
):  # type: (NeuralNetworkBuilder, Node, Graph, ErrorHandling) -> None
    """
    convert to CoreML BatchNorm Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L1633
    """
    epsilon = node.attrs.get("epsilon", 1e-5)
    if (
        node.inputs[1] not in node.input_tensors
        or node.inputs[2] not in node.input_tensors
    ):
        return err.unsupported_op_configuration(
            builder,
            node,
            graph,
            "CoreML InstanceNorm requires Scale and Bias to be known",
        )

    scale = node.input_tensors[node.inputs[1]]
    bias = node.input_tensors[node.inputs[2]]

    rank = builder._get_rank(node.inputs[0])
    # ONNX converts B x C tensor into B x C x 1 hence
    # Rank 2 BN is mapped to Rank 3 BN
    if rank == 3:
        # 1D Batch Norm
        add_bn_with_expansion(
            builder,
            node,
            err,
            node.name,
            node.inputs[0],
            node.outputs[0],
            scale.shape[0],
            scale,
            bias,
            epsilon=epsilon,
            compute_mean_var=True,
            instance_normalization=True,
            axes_for_expansion=[0, 3],
        )
    elif rank == 4:
        # 2D Batch Norm
        add_bn_with_expansion(
            builder,
            node,
            err,
            node.name,
            node.inputs[0],
            node.outputs[0],
            scale.shape[0],
            scale,
            bias,
            epsilon=epsilon,
            compute_mean_var=True,
            instance_normalization=True,
            axes_for_expansion=[],
        )
    else:
        # Unsupported 1D, 3D and above
        err.unsupported_op_configuration(
            builder, node, graph, "provided number axes {} not supported".format(rank)
        )


def _convert_less(builder, node, graph, err):
    """
    convert to CoreML Less Than Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L907
    """
    load_input_constants(builder, node, graph, err)
    builder.add_less_than(
        name=node.name, input_names=node.inputs, output_name=node.outputs[0],
    )


def _convert_lstm(
    builder, node, graph, err
):  # type: (NeuralNetworkBuilder, Node, Graph, ErrorHandling) -> None
    """
    convert to CoreML Uni/Bi-Directional LSTM Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3282
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3348
    """

    def get_weights(W, W_name, R, R_name, B):
        """
        Helper routine to return weights in CoreML LSTM required format
        """
        W = np.expand_dims(np.expand_dims(W, 3), 3)
        R = np.expand_dims(np.expand_dims(R, 3), 3)

        if W is None:
            err.missing_initializer(
                node,
                "Weight tensor: {} not found in the graph initializer".format(W_name),
            )
        if R is None:
            err.missing_initializer(
                node,
                "Weight tensor: {} not found in the graph initializer".format(R_name),
            )

        W_i, W_o, W_f, W_c = np.split(np.squeeze(W), 4)  # type: ignore
        R_i, R_o, R_f, R_c = np.split(np.squeeze(R), 4)  # type: ignore

        W_x = [W_i, W_f, W_o, W_c]
        W_h = [R_i, R_f, R_o, R_c]
        b = None
        if B is not None:
            b_Wi, b_Wo, b_Wf, b_Wc, b_Ri, b_Ro, b_Rf, b_Rc = np.split(np.squeeze(B), 8)  # type: ignore
            b = [b_Wi + b_Ri, b_Wf + b_Rf, b_Wo + b_Ro, b_Wc + b_Rc]

        return W_x, W_h, b

    def expand_dim(node_name, input_name, output_name, axes):
        builder.add_expand_dims(
            name=node_name, input_name=input_name, output_name=output_name, axes=axes
        )

    # Read attributes
    # activation alpha and beta
    if "activation_alpha" in node.attrs or "activation_beta" in node.attrs:
        err.unsupported_feature_warning(
            node, "Activation parameter alpha and beta are currently not used"
        )

    inner_activation = "SIGMOID"
    cell_state_update_activation = "TANH"
    output_activation = "TANH"

    if "activations" in node.attrs:
        activations_list = node.attrs["activations"]

        if len(activations_list) < 3:
            err.unsupported_op_configuration(
                builder,
                node,
                graph,
                "Error in ONNX model: Less number of activations provided",
            )

        if len(activations_list) == 6:
            err.unsupported_feature_warning(
                node, "Forward and backward pass will use same activations."
            )

        inner_activation = activations_list[0].upper()
        cell_state_update_activation = activations_list[1].upper()
        output_activation = activations_list[2].upper()

    # Provide max Clip Value if not provided
    clip_threshold = node.attrs.get("clip", 500000.0)

    # Extract direction from ONNX attribute
    direction = 1
    if (
        "direction" in node.attrs
        and node.attrs["direction"].decode("utf-8") == "bidirectional"
    ):
        direction = 2

    hidden_size = node.attrs.get("hidden_size")

    input_forget = node.attrs.get("input_forget", 0) == 1

    # Read inputs
    W_name = node.inputs[1]
    R_name = node.inputs[2]
    B = None
    if len(node.inputs) > 3:
        B_name = node.inputs[3]
        B = node.input_tensors.get(B_name, None)

    W = node.input_tensors.get(W_name, None)
    R = node.input_tensors.get(R_name, None)

    W = np.split(W, direction)
    R = np.split(R, direction)
    if B is not None:
        B = np.split(B, direction)
    else:
        B = [None, None]

    # Get weights for forward direction
    W_x, W_h, b = get_weights(W[0], W_name, R[0], R_name, B[0])

    # shape of input
    input_size = W_x[0].shape[1]

    # Get input and output for hidden and cell
    input_h = node.inputs[5] if len(node.inputs) > 5 else node.inputs[0] + "_h_input"
    input_c = node.inputs[6] if len(node.inputs) > 6 else node.inputs[0] + "_c_input"
    output_h = (
        node.outputs[1] if len(node.outputs) > 1 else node.outputs[0] + "_h_output"
    )
    output_c = (
        node.outputs[2] if len(node.outputs) > 2 else node.outputs[0] + "_c_output"
    )
    output_h_5d = output_h + "_5d"
    output_c_5d = output_c + "_5d"

    # if input is not present in the network, load they as constant
    load_input_constants(builder, node, graph, err)

    # Input is represented as [Seq Len, Batch Size, Input Size]
    if len(node.inputs) < 6:
        batch_size = graph.shape_dict[node.inputs[0]][1]
        builder.add_load_constant_nd(
            name=node.name + "_load_initial_h_and_c",
            output_name=input_h,
            constant_value=0.0,
            shape=[direction, batch_size, hidden_size],
        )
        # OPTIMIZATION: let's reuse the intial weights
        input_c = input_h

    # Get tensors for peepholes
    peepholes = node.inputs[7] if len(node.inputs) > 7 else None

    # CoreML LSTM expects 5-d tensor
    # Expand dimensions of input to 5-d for compatibility
    rank = builder._get_rank(node.inputs[0])
    if rank == -1:
        return err.unsupported_op_configuration(
            builder, node, graph, "Rank unknown for input"
        )
    if rank < 5:
        add_nodes = 5 - rank
        # TODO: Add one expand instead of adding one after another for input, h and c
        expand_dim(
            node.name + "_expand_in_0",
            node.inputs[0],
            node.inputs[0] + "_expand_out_0",
            [rank],
        )
        expand_dim(
            node.name + "_expand_in_h_0", input_h, input_h + "_expand_out_h_0", [rank]
        )
        expand_dim(
            node.name + "_expand_in_c_0", input_c, input_c + "_expand_out_c_0", [rank]
        )

        for i in range(1, add_nodes):
            i_str = str(i)
            i_p_str = str(i - 1)
            expand_dim(
                node.name + "_expand_in_" + i_str,
                node.inputs[0] + "_expand_out_" + i_p_str,
                node.inputs[0] + "_expand_out_" + i_str,
                [rank + i],
            )
            expand_dim(
                node.name + "_expand_in_h_" + i_str,
                input_h + "_expand_out_h_" + i_p_str,
                input_h + "_expand_out_h_" + i_str,
                [rank + i],
            )
            expand_dim(
                node.name + "_expand_in_c_" + i_str,
                input_c + "_expand_out_c_" + i_p_str,
                input_c + "_expand_out_c_" + i_str,
                [rank + i],
            )

    if direction == 1:
        # Peephole from ONNX are of shape [Num Dir, 3 * hidden_size]
        # Reshape into CoreML format of [input hs, forget hs, cell hs]
        if peepholes is not None:
            builder.add_reshape_static(
                name=node.name + "_peephole_reshape",
                input_name=peepholes,
                output_name=peepholes + "_reshaped",
                output_shape=[hidden_size, hidden_size, hidden_size],
            )
            peepholes = peepholes + "_reshaped"

        builder.add_unilstm(
            name=node.name,
            W_h=W_h,
            W_x=W_x,
            b=b,
            hidden_size=hidden_size,
            input_size=input_size,
            input_names=[
                node.inputs[0] + "_expand_out_" + str(add_nodes - 1),
                input_h + "_expand_out_h_" + str(add_nodes - 1),
                input_c + "_expand_out_c_" + str(add_nodes - 1),
            ],
            output_names=[node.outputs[0] + "_5d_out", output_h_5d, output_c_5d],
            inner_activation=inner_activation,
            cell_state_update_activation=cell_state_update_activation,
            output_activation=output_activation,
            peep=peepholes,
            output_all=True,
            forget_bias=True,
            coupled_input_forget_gate=input_forget,
            cell_clip_threshold=clip_threshold,
            reverse_input=False,
        )
    elif direction == 2:
        if len(W) != 2 and len(R) != 2 and len(B) != 2:
            err.unsupported_op_configuration(
                builder,
                node,
                graph,
                "Bi-Directional LSTM does not have weights for both the directions",
            )

        W_x_back, W_h_back, b_back = get_weights(W[1], W_name, R[1], R_name, B[1])

        peephole_f = None
        peephole_b = None
        if peepholes is not None:
            builder.add_reshape_static(
                name=node.name + "_peephole_reshape",
                input_name=peepholes,
                output_name=peepholes + "_reshaped",
                output_shape=[direction, hidden_size, hidden_size, hidden_size],
            )

            peepholes_f = peepholes + "_f"
            peepholes_b = peepholes + "_b"

            builder.add_split_nd(
                name=node.name + "_peephole_split",
                input_name=peepholes + "_reshaped",
                output_names=[peepholes_f, peepholes_b],
                axis=0,
            )

        # split input_h and input_c into two parts
        builder.add_split_nd(
            name=node.name + "_split_h",
            input_name=input_h + "_expand_out_h_" + str(add_nodes - 1),
            output_names=[input_h + "_f", input_h + "_b"],
            axis=0,
        )

        # OPTIMIZATION: If input_h and input_c are same
        # Avoid creating new split and instead reuse
        if input_h != input_c:
            builder.add_split_nd(
                name=node.name + "_split_c",
                input_name=input_c + "_expand_out_c_" + str(add_nodes - 1),
                output_names=[input_c + "_f", input_c + "_b"],
                axis=0,
            )

        builder.add_bidirlstm(
            name=node.name,
            W_h=W_h,
            W_x=W_x,
            b=b,
            W_h_back=W_h_back,
            W_x_back=W_x_back,
            b_back=b_back,
            hidden_size=hidden_size,
            input_size=input_size,
            input_names=[
                node.inputs[0] + "_expand_out_" + str(add_nodes - 1),
                input_h + "_f",
                input_c + "_f",
                input_h + "_b",
                input_c + "_b",
            ],
            output_names=[
                node.outputs[0] + "_5d_out",
                output_h + "_f",
                output_c + "_f",
                output_h + "_b",
                output_c + "_b",
            ],
            inner_activation=inner_activation,
            cell_state_update_activation=cell_state_update_activation,
            output_activation=output_activation,
            output_all=True,
            peep=peephole_f,
            peep_back=peephole_b,
            forget_bias=True,
            coupled_input_forget_gate=input_forget,
            cell_clip_threshold=clip_threshold,
        )

        # Combine output_h and output_c
        builder.add_concat_nd(
            name=node.name + "concat_output_h",
            input_names=[output_h + "_f", output_h + "_b"],
            output_name=output_h_5d,
            axis=0,
        )

        builder.add_concat_nd(
            name=node.name + "concat_output_c",
            input_names=[output_c + "_f", output_c + "_b"],
            output_name=output_c_5d,
            axis=0,
        )
    else:
        err.unsupported_op_configuration(
            builder, node, graph, "Unsupported direction {} for LSTM".format(direction)
        )

    # CoreML output is [Seq Len, Batch Size, Num Dir * Hidden Size, 1, 1]
    # Return output as [Seq Len, Num Dir, Batch Size, Hidden Size]
    # Following steps:
    #       a. Reshape and split hidden size for direction [Seq Len, Batch Size, Num Dir, Hidden Size, 1]
    #       b. Squeeze last dimension [Seq Len, Batch Size, Num Dir, Hidden Size]
    #       c. Permute to fix the order [Seq Len, Num Dir, Batch Size, Hidden Size, 1]
    builder.add_rank_preserving_reshape(
        name=node.name + "_reshape_",
        input_name=node.outputs[0] + "_5d_out",
        output_name=node.outputs[0] + "_5d_reshaped",
        output_shape=[0, 0, direction, -1, 0],
    )

    builder.add_squeeze(
        name=node.name + "_squeeze_out",
        input_name=node.outputs[0] + "_5d_reshaped",
        output_name=node.outputs[0] + "_4d",
        axes=[-1],
    )

    builder.add_transpose(
        name=node.name + "_transpose",
        axes=[0, 2, 1, 3],
        input_name=node.outputs[0] + "_4d",
        output_name=node.outputs[0],
    )

    # Squeeze dimensions of output_h and output_c
    builder.add_squeeze(
        name=node.name + "_squeeze_out_h",
        input_name=output_h_5d,
        output_name=output_h,
        axes=[-1, -2],
    )
    builder.add_squeeze(
        name=node.name + "_squeeze_out_c",
        input_name=output_c_5d,
        output_name=output_c,
        axes=[-1, -2],
    )


def _convert_logical(builder, node, graph, err):
    """
    convert to CoreML Logical And/Or/Xor/Not Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L1013
    """
    mode = node.op_type.upper()
    builder.add_logical(
        name=node.name, input_names=node.inputs, output_name=node.outputs[0], mode=mode
    )


def _convert_pad(builder, node, graph, err):
    """
    convert to CoreML Padding / ConstantPadding Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4397
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L1822
    """
    mode = node.attrs.get("mode", "constant")

    try:
        mode = mode.decode()
    except (UnicodeDecodeError, AttributeError):
        pass

    if mode == "constant":
        pads = node.attrs.get("pads", [])
        value = node.attrs.get("value", 0.0)
        # onnx padding spec: [x1_top, ..., xn_top, x1_bottom, ..., xn_bottom]
        # coreml padding spec: [x1_top, x1_bottom, ..., xn_top, xn_bottom]
        assert len(pads) % 2 == 0, 'even number of pads expected'
        pads_coreml = [None] * len(pads)
        pads_coreml[::2] = pads[:len(pads) // 2]
        pads_coreml[1::2] = pads[len(pads) // 2:]

        builder.add_constant_pad(
            name=node.name,
            input_names=node.inputs,
            output_name=node.outputs[0],
            value=value,
            pad_to_given_output_size_mode=False,
            pad_amounts=pads_coreml,
        )
    else:
        _convert_pad_5d(builder, node, graph, err)


def _convert_matmul(builder, node, graph, err):
    """
    convert to CoreML BatchedMatMul Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3473
    """
    weight_name = node.inputs[1]
    W = None
    weight_as_layer_parameter = False
    if weight_name in node.input_tensors:
        W = node.input_tensors[weight_name]

    if W is not None:
        if len(W.shape) != 2:
            # since weight as parameter in batchedMatMul layer must be rank 2
            builder.add_load_constant_nd(
                node.name + "_const_weight_input",
                weight_name,
                constant_value=W,
                shape=W.shape,
            )
        else:
            weight_as_layer_parameter = True

    if weight_as_layer_parameter:
        builder.add_batched_mat_mul(
            name=node.name,
            input_names=[node.inputs[0]],
            output_name=node.outputs[0],
            weight_matrix_rows=W.shape[0],
            weight_matrix_columns=W.shape[1],
            W=W,
        )
    else:
        builder.add_batched_mat_mul(
            name=node.name,
            input_names=[node.inputs[0], weight_name],
            output_name=node.outputs[0],
        )


def _convert_max(builder, node, graph, err):
    """
    convert to CoreML Max Broadcastable Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4126
    """
    load_input_constants(builder, node, graph, err)
    add_broadcastable_op_chain(builder, node, err, builder.add_max_broadcastable)


def _convert_mean(builder, node, graph, err):
    """
    convert to CoreML Add Broadcastable Layer and Divide BroadCastable Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4117
    """
    number_of_inputs = len(node.inputs)
    output_name = node.outputs[0]
    node.outputs[0] = node.outputs[0] + "_sum"

    builder.add_load_constant_nd(
        name=node.name + "_divider",
        output_name=output_name + "_divider",
        constant_value=np.array(number_of_inputs),
        shape=[1],
    )
    add_broadcastable_op_chain(builder, node, err, builder.add_add_broadcastable)
    builder.add_divide_broadcastable(
        name=node.name + "_mean",
        input_names=[node.outputs[0], output_name + "_divider"],
        output_name=output_name,
    )


def _convert_pow(builder, node, graph, err):
    """
    convert to CoreML Pow Broadcastable Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3969
    """
    load_input_constants(builder, node, graph, err)
    add_broadcastable_op_chain(builder, node, err, builder.add_pow_broadcastable)


def _convert_randomnormal(builder, node, graph, err):
    """
    convert to CoreML Random Normal Static Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4457
    """
    add_random(builder, node, graph, err, builder.add_random_normal_static)


def _convert_randomnormallike(builder, node, graph, err):
    """
    convert to CoreML Random Normal Like Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4434
    """
    # Ignoring attribute `dtype` as CoreML internally represents tensors into 'Float'
    mean = node.attributes.get("mean", 0.0)
    scale = node.attributes.get("scale", 1.0)
    seed = node.attributes.get("seed", -1)

    builder.add_random_normal_like(
        name=node.name,
        input_name=node.inputs[0],
        output_name=node.outputs[0],
        mean=mean,
        stddev=scale,
        seed=seed,
    )


def _convert_randomuniform(builder, node, graph, err):
    """
    convert to CoreML Random Uniform Static Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4526
    """
    add_random(builder, node, graph, err, builder.random_uniform_static)


def _convert_randomuniformlike(builder, node, graph, err):
    """
    convert to CoreML Random Normal Like Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4503
    """
    # Ignoring attribute `dtype` as CoreML internally represents tensors into 'Float'
    mean = node.attributes.get("mean", 0.0)
    scale = node.attributes.get("scale", 1.0)
    seed = node.attributes.get("seed", -1)

    builder.add_random_uniform_like(
        name=node.name,
        input_name=node.inputs[0],
        output_name=node.outputs[0],
        mean=mean,
        stddev=scale,
        seed=seed,
    )


def _convert_min(builder, node, graph, err):
    """
    convert to CoreML Min Broadcastable Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4135
    """
    load_input_constants(builder, node, graph, err)
    add_broadcastable_op_chain(builder, node, err, builder.add_min_broadcastable)


def _convert_mod(builder, node, graph, err):
    """
    convert to CoreML Mod Broadcastable Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4144
    """
    load_input_constants(builder, node, graph, err)
    add_broadcastable_op_chain(builder, node, err, builder.add_mod_broadcastable)


def _convert_mul(builder, node, graph, err):
    """
    convert to CoreML Multiply Broadcastable Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4171
    """
    load_input_constants(builder, node, graph, err)
    add_broadcastable_op_chain(builder, node, err, builder.add_multiply_broadcastable)


def _convert_nonzero(builder, node, graph, err):
    """
    convert to CoreML Where Non Zero Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4002
    """
    load_input_constants(builder, node, graph, err)
    builder.add_where_nonzero(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_pool(builder, node, graph, err):
    """
    convert to CoreML Pooling Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L477
    """
    params_dict = dict()
    params_dict["is_global"] = False
    if node.op_type.startswith("Global"):
        params_dict["is_global"] = True
    if node.op_type.endswith("MaxPool"):
        params_dict["layer_type"] = "MAX"
    elif node.op_type.endswith("AveragePool"):
        params_dict["layer_type"] = "AVERAGE"
    else:
        return err.unsupported_op_configuration(
            builder, node, graph, "Unsupported pool type"
        )

    if len(node.outputs) == 2:
        return err.unsupported_op_configuration(
            builder, node, graph, "argmax with pool unsupported"
        )

    if "ceil_mode" in node.attrs and node.attrs["ceil_mode"] == 1:
        return err.unsupported_op_configuration(
            builder, node, graph, "ceil_mode=1 not supported"
        )

    if "dilations" in node.attrs:
        return err.unsupported_op_configuration(
            builder, node, graph, "dilations not supported"
        )

    _add_conv_like_op(
        _add_pool, _get_pool_params, params_dict, builder, node, graph, err
    )


def _convert_reduce(builder, node, graph, err):
    """
    convert to CoreML ReduceSum Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4707
    """
    load_input_constants(builder, node, graph, err)

    # read attributes
    axes = node.attrs.get("axes", None)
    reduce_all = False
    if axes is None:
        reduce_all = True
    keepdims = node.attrs.get("keepdims", True)

    # add respective operator
    op_type = node.op_type
    if op_type == "ReduceSum":
        builder.add_reduce_sum(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            axes=axes,
            keepdims=keepdims,
            reduce_all=reduce_all,
        )
    elif op_type == "ReduceProd":
        builder.add_reduce_prod(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            axes=axes,
            keepdims=keepdims,
            reduce_all=reduce_all,
        )
    elif op_type == "ReduceMean":
        builder.add_reduce_mean(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            axes=axes,
            keepdims=keepdims,
            reduce_all=reduce_all,
        )
    elif op_type == "ReduceMax":
        builder.add_reduce_max(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            axes=axes,
            keepdims=keepdims,
            reduce_all=reduce_all,
        )
    elif op_type == "ReduceMin":
        builder.add_reduce_min(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            axes=axes,
            keepdims=keepdims,
            reduce_all=reduce_all,
        )
    elif op_type == "ReduceL2":
        builder.add_reduce_l2(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            axes=axes,
            keepdims=keepdims,
            reduce_all=reduce_all,
        )
    elif op_type == "ReduceL1":
        builder.add_reduce_l1(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            axes=axes,
            keepdims=keepdims,
            reduce_all=reduce_all,
        )
    elif op_type == "ReduceSumSquare":
        builder.add_reduce_sumsquare(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            axes=axes,
            keepdims=keepdims,
            reduce_all=reduce_all,
        )
    elif op_type == "ReduceLogSum":
        builder.add_reduce_logsum(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            axes=axes,
            keepdims=keepdims,
            reduce_all=reduce_all,
        )
    elif op_type == "ReduceLogSumExp":
        builder.add_reduce_logsumexp(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            axes=axes,
            keepdims=keepdims,
            reduce_all=reduce_all,
        )
    else:
        err.unsupported_op_configuration(
            builder, node, graph, "Unsupported reduce operation: {}".format(op_type)
        )


def _convert_reshape(builder, node, graph, err):
    """
    convert to CoreML Reshape Static Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4844
    """
    shape_node = node.inputs[1]
    if shape_node in node.input_tensors:
        output_shape = node.input_tensors[shape_node].astype(np.int64)

        # if rank is same, then call rank preserving reshape
        if node.inputs[0] not in graph.shape_dict:
            # If Input shape is not present and output shape is known
            # add reshape static as
            # TODO: ONNX should be able to infer the shape
            builder.add_reshape_static(
                name=node.name,
                input_name=node.inputs[0],
                output_name=node.outputs[0],
                output_shape=output_shape,
            )
            return

        len_of_input_shape = builder._get_rank(node.inputs[0])
        if len(output_shape) == len_of_input_shape:
            builder.add_rank_preserving_reshape(
                name=node.name,
                input_name=node.inputs[0],
                output_name=node.outputs[0],
                output_shape=output_shape,
            )
        else:
            add_static_reshape = True
            if len_of_input_shape > len(output_shape):
                # Output rank is less than input rank
                # Case when most of the dims size is unchanged
                num_zeros = 0
                num_neg_ones = 0
                for i in output_shape:
                    if i == 0:
                        num_zeros += 1
                    elif i == -1:
                        num_neg_ones += 1

                if num_neg_ones > 1:
                    err.unsupported_op_configuration(
                        builder,
                        node,
                        graph,
                        "Error in ONNX model: At most one dimension of new shape can be -1, found {}".format(
                            num_neg_ones
                        ),
                    )

                if num_neg_ones + num_zeros == len(output_shape):
                    # Rank of output is less than input
                    # Make Rank equivalent for reshape and then squeeze
                    add_static_reshape = False
                    new_shape = []
                    i = 0
                    for i in range(len(output_shape)):
                        new_shape.append(output_shape[i])
                        if output_shape[i] == -1:
                            break
                    while i < len_of_input_shape - 1:
                        new_shape.append(1)
                        i += 1

                    builder.add_rank_preserving_reshape(
                        name=node.name + "_reshape_preserving",
                        input_name=node.inputs[0],
                        output_name=node.outputs[0] + "_reshape_dim_preserved",
                        output_shape=new_shape,
                    )

                    squeeze_axes = list(
                        range(len(output_shape) - len_of_input_shape, 0)
                    )
                    squeeze_axes.reverse()

                    builder.add_squeeze(
                        name=node.name,
                        input_name=node.outputs[0] + "_reshape_dim_preserved",
                        output_name=node.outputs[0],
                        axes=squeeze_axes,
                    )

            if add_static_reshape:
                builder.add_reshape_static(
                    name=node.name,
                    input_name=node.inputs[0],
                    output_name=node.outputs[0],
                    output_shape=output_shape,
                )
    else:
        builder.add_reshape_dynamic(
            name=node.name, input_names=node.inputs, output_name=node.outputs[0],
        )


def _convert_resize(builder, node, graph, err):
    """
    convert to CoreML Upsample or Resize Bilinear Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L2139
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L2178
    """
    mode = node.attrs.get("mode", "nearest")
    if node.inputs[1] not in node.input_tensors:
        return err.unsupported_op_configuration(
            builder,
            node,
            graph,
            "Scaling factor unknown!! CoreML does not support dynamic scaling for Resize",
        )

    mode = "NN" if mode == "nearest" else "BILINEAR"
    scale = node.input_tensors[node.inputs[1]]

    builder.add_upsample(
        name=node.name,
        scaling_factor_h=scale[-2],
        scaling_factor_w=scale[-1],
        input_name=node.inputs[0],
        output_name=node.outputs[0],
        mode=mode,
    )


def _convert_reverse_sequence(builder, node, graph, err):
    """
    convert to CoreML Reverse Sequence Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3577
    """
    batch_axis = node.attrs.get("batch_axis", 1)
    time_axis = node.attrs.get("time_axis", 0)

    output_name = node.outputs[0]
    add_transpose = False
    if batch_axis > time_axis:
        output_name += "_before_reverse"
        batch_axis, time_axis = time_axis, batch_axis
        add_transpose = True

    builder.add_reverse_sequence(
        name=node.name,
        input_names=node.inputs,
        output_name=output_name,
        batch_axis=batch_axis,
        seq_axis=time_axis,
    )

    if add_transpose:
        output_name_post = "_before_reverse"
        rank = builder._get_rank(node.inputs[0])
        if rank == -1:
            return err.unsupported_op_configuration(
                builder, node, graph, "Rank unknown for input"
            )
        axes = list(range(rank))
        axes[batch_axis], axes[time_axis] = axes[time_axis], axes[batch_axis]
        builder.add_transpose(
            name=node.name + "_transpose",
            axes=axes,
            input_name=output_name,
            output_name=node.outputs[0],
        )


def _convert_roialign(builder, node, graph, err):
    """
    convert to CoreML CropResize and Pooling Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L2239
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L1702
    """

    target_height = node.attrs.get("output_height", 1)
    target_width = node.attrs.get("output_width", 1)
    mode = node.attrs.get("mode", "AVERAGE").upper()
    sampling_ratio = node.attrs.get("sampling_ratio", 0)
    spatial_scale = node.attrs.get("sampling_scale", 1.0)

    if node.inputs[2] in graph.inputs:
        graph.inputs.remove(node.inputs[2])

    builder.add_expand_dims(
        name=node.name + "_expand_0",
        input_name=node.inputs[0],
        output_name=node.inputs[0] + "_expanded",
        axes=[0],
    )
    node.inputs[0] += "_expanded"

    builder.add_expand_dims(
        name=node.name + "_expand_2",
        input_name=node.inputs[2],
        output_name=node.inputs[2] + "_expanded",
        axes=[1],
    )
    node.inputs[2] += "_expanded"

    builder.add_concat_nd(
        name=node.name + "_concat_indices",
        input_names=[node.inputs[2], node.inputs[1]],
        output_name=node.inputs[1] + "_rois",
        axis=1,
    )
    node.inputs[1] += "_rois"

    builder.add_expand_dims(
        name=node.name + "_expand_1",
        input_name=node.inputs[1],
        output_name=node.inputs[1] + "_expanded",
        axes=[1, 3, 4],
    )
    node.inputs[1] += "_expanded"

    builder.add_crop_resize(
        name=node.name + "_crop_resize",
        input_names=[node.inputs[0], node.inputs[1]],
        output_name=node.outputs[0] + "_crop_resized",
        target_height=target_height * sampling_ratio,
        target_width=target_width * sampling_ratio,
        mode="ROI_ALIGN_MODE",
        box_indices_mode="CORNERS_WIDTH_FIRST",
        spatial_scale=spatial_scale,
    )

    builder.add_squeeze(
        name=node.name + "_squeeze",
        input_name=node.outputs[0] + "_crop_resized",
        output_name=node.outputs[0] + "_crop_resized_squeezed",
        axes=[1],
    )

    builder.add_pooling(
        name=node.name + "_pool",
        height=sampling_ratio,
        width=sampling_ratio,
        layer_type=mode,
        input_name=node.outputs[0] + "_crop_resized_squeezed",
        output_name=node.outputs[0],
        stride_height=sampling_ratio,
        stride_width=sampling_ratio,
        padding_type="VALID",
    )


def _convert_round(builder, node, graph, err):
    """
    convert to CoreML Round Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L5029
    """
    builder.add_round(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_scatter(builder, node, graph, err):
    """
    convert to CoreML Scatter Along Axis Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4308
    """
    axis = node.attrs.get("axis", 0)
    builder.add_scatter_along_axis(
        name=node.name, input_names=node.inputs, output_name=node.outputs[0], axis=axis
    )


def _convert_size(builder, node, graph, err):
    """
    convert to CoreML GetShape and ReduceProd Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L5131
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4722
    """
    builder.add_get_shape(
        name=node.name,
        input_name=node.inputs[0],
        output_name=node.inputs[0] + "_getshape",
    )
    builder.add_reduce_prod(
        name=node.name + "_reduce_prod",
        input_name=node.inputs[0] + "_getshape",
        output_name=node.outputs[0],
    )


def _convert_slice_ir4v9(builder, node, graph, err):
    """
    convert to CoreML Slice Static Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L5082
    """
    if node.inputs[0] in graph.shape_dict:
        data_shape = graph.shape_dict[node.inputs[0]]
    else:
        rank = builder._get_rank(node.inputs[0])
        if rank == -1:
            return err.unsupported_op_configuration(
                builder, node, graph, "Input shape not available"
            )
        data_shape = [INT_MAX] * rank

    len_of_data = len(data_shape)
    begin_masks = [True] * len_of_data
    end_masks = [True] * len_of_data

    default_axes = list(range(len_of_data))
    default_steps = [1] * len_of_data

    ip_starts = node.attrs.get("starts")
    ip_ends = node.attrs.get("ends")
    axes = node.attrs.get("axes", default_axes)
    steps = node.attrs.get("steps", default_steps)

    starts = [0] * len_of_data
    ends = [0] * len_of_data

    for i in range(len(axes)):
        current_axes = axes[i]
        starts[current_axes] = ip_starts[i]
        ends[current_axes] = ip_ends[i]
        # n <= end <= INT_MAX implies end is -1, hence end_mask should be True
        # otherwise end_mask should be False
        if ends[current_axes] < data_shape[current_axes]:
            # this means end is not -1
            end_masks[current_axes] = False

        if starts[current_axes] != 0:
            begin_masks[current_axes] = False

    builder.add_slice_static(
        name=node.name,
        input_name=node.inputs[0],
        output_name=node.outputs[0],
        begin_ids=starts,
        end_ids=ends,
        strides=steps,
        begin_masks=begin_masks,
        end_masks=end_masks,
    )


def _convert_slice(builder, node, graph, err):
    """
    convert to CoreML Slice Static Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L5082
    """
    if len(node.inputs) == 1:
        return _convert_slice_ir4v9(builder, node, graph, err)

    if node.inputs[0] not in graph.shape_dict:
        err.unsupported_op_configuration(
            builder, node, graph, "Input shape not available"
        )

    data_shape = graph.shape_dict[node.inputs[0]]
    len_of_data = len(data_shape)
    begin_masks = [True] * len_of_data
    end_masks = [True] * len_of_data

    default_axes = list(range(len_of_data))

    add_static_slice_layer = False
    if node.inputs[1] in node.input_tensors and node.inputs[2] in node.input_tensors:
        if len(node.inputs) > 3:
            if node.inputs[3] in node.input_tensors:
                if len(node.inputs) > 4:
                    if node.inputs[4] in node.input_tensors:
                        add_static_slice_layer = True
                else:
                    add_static_slice_layer = True
        else:
            add_static_slice_layer = True

    if add_static_slice_layer:
        ip_starts = node.input_tensors[node.inputs[1]]
        ip_ends = node.input_tensors[node.inputs[2]]
        axes = (
            node.input_tensors[node.inputs[3]] if len(node.inputs) > 3 else default_axes
        )
        ip_steps = node.input_tensors[node.inputs[4]] if len(node.inputs) > 4 else None

        starts = [0] * len_of_data
        ends = [0] * len_of_data
        steps = [1] * len_of_data

        for i in range(len(axes)):
            current_axes = axes[i]
            starts[current_axes] = ip_starts[i]
            ends[current_axes] = ip_ends[i]
            # n <= end <= INT_MAX implies end is -1, hence end_mask should be True
            # otherwise end_mask should be False
            if ends[current_axes] < data_shape[current_axes]:
                # this means end is not -1
                end_masks[current_axes] = False

            if starts[current_axes] != 0:
                begin_masks[current_axes] = False

            if isinstance(ip_steps, list):
                steps[current_axes] = ip_steps[i]

        builder.add_slice_static(
            name=node.name,
            input_name=node.inputs[0],
            output_name=node.outputs[0],
            begin_ids=starts,
            end_ids=ends,
            strides=steps,
            begin_masks=begin_masks,
            end_masks=end_masks,
        )
    else:
        err.unsupported_op_configuration(
            builder,
            node,
            graph,
            "CoreML does not support Dynamic Slice with unknown axes. Please provide Custom Function/Layer",
        )


def _convert_softmax_nd(builder, node, graph, err):
    """
    convert to CoreML SoftMax ND Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#3547
    """
    axis = node.attrs.get("axis", 1)
    builder.add_softmax_nd(
        name=node.name,
        input_name=node.inputs[0],
        output_name=node.outputs[0]
        + ("_softmax" if node.op_type == "LogSoftmax" else ""),
        axis=axis,
    )
    if node.op_type == "LogSoftmax":
        builder.add_unary(
            name=node.name + "_log",
            input_name=node.outputs[0] + "_softmax",
            output_name=node.outputs[0],
            mode="log",
        )


def _convert_softmax(builder, node, graph, err):
    """
    convert to CoreML SoftMax ND Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#3547
    """

    def add_softmax(output_name, rank=-1, axis=-3):
        softmax_axis = 3
        axes = list(range(5 - rank))
        if axis < 0:
            axis = rank + axis
        axis += len(axes)
        softmax_output_name = output_name + "_expanded"

        expanded_node = node.name + "_" + node.inputs[0] + "_expanded"
        builder.add_expand_dims(
            name=node.name + "_expand_dims",
            input_name=node.inputs[0],
            output_name=expanded_node,
            axes=axes,
        )
        input_name = expanded_node
        rank = 5

        if axis != -3 and axis != rank - softmax_axis:
            transpose_axes = list(range(rank))
            transpose_axes[-3], transpose_axes[axis] = (
                transpose_axes[axis],
                transpose_axes[-3],
            )

            builder.add_transpose(
                name=node.name + "_transpose",
                axes=transpose_axes,
                input_name=input_name,
                output_name=input_name + "_transposed",
            )
            input_name += "_transposed"
            softmax_output_name += "_transposed"

        builder.add_softmax(
            name=node.name, input_name=input_name, output_name=softmax_output_name
        )

        if axis != -3 and axis != rank - softmax_axis:
            transpose_axes = list(range(rank))
            transpose_axes[-3], transpose_axes[axis] = (
                transpose_axes[axis],
                transpose_axes[-3],
            )

            builder.add_transpose(
                name=node.name + "_transpose_back",
                axes=transpose_axes,
                input_name=softmax_output_name,
                output_name=softmax_output_name + "_transposed_back",
            )
            softmax_output_name += "_transposed_back"

        builder.add_squeeze(
            name=node.name + "_squeeze_dims",
            input_name=softmax_output_name,
            output_name=output_name,
            axes=axes,
        )

    axis = node.attrs.get("axis", 1)
    rank = builder._get_rank(node.inputs[0])
    if rank == -1:
        return _convert_softmax_nd(builder, node, graph, err)

    if node.op_type == "LogSoftmax":
        add_softmax(node.outputs[0] + "_softmax", rank=rank, axis=axis)
        builder.add_unary(
            name=node.name + "_log",
            input_name=node.outputs[0] + "_softmax",
            output_name=node.outputs[0],
            mode="log",
        )
    else:
        add_softmax(node.outputs[0], rank=rank, axis=axis)


def _convert_split(builder, node, graph, err):
    """
    convert to CoreML Squeeze Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#5003
    """
    axis = node.attrs.get("axis", 0)
    split = node.attrs.get("split", None)
    num_splits = len(node.outputs) if split is None else 2

    builder.add_split_nd(
        name=node.name,
        input_name=node.inputs[0],
        output_names=node.outputs,
        axis=axis,
        num_splits=num_splits,
        split_sizes=split,
    )


def _convert_shape(builder, node, graph, err):
    """
    convert to CoreML GetShape Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L5131
    """
    builder.add_get_shape(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_squeeze(builder, node, graph, err):
    """
    convert to CoreML Squeeze Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4903
    """
    axes = node.attrs.get("axes", None)
    builder.add_squeeze(
        name=node.name,
        input_name=node.inputs[0],
        output_name=node.outputs[0],
        axes=axes,
    )


def _convert_sub(builder, node, graph, err):
    """
    convert to CoreML Subtract Broadcastable Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4117
    """
    load_input_constants(builder, node, graph, err)
    add_broadcastable_op_chain(builder, node, err, builder.add_subtract_broadcastable)


def _convert_tanh(builder, node, graph, err):
    """
    convert to CoreML Tanh Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3881
    """
    load_input_constants(builder, node, graph, err)
    builder.add_tanh(
        name=node.name, input_name=node.inputs[0], output_name=node.outputs[0]
    )


def _convert_tile(builder, node, graph, err):
    """
    convert to CoreML Tile Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L5117
    """
    load_input_constants(builder, node, graph, err)
    if node.inputs[1] not in node.input_tensors:
        err.unsupported_op_configuration(
            builder,
            node,
            graph,
            "CoreML Tile layer does not support dynamic 'reps'. 'reps' should be known statically",
        )
    builder.add_tile(
        name=node.name,
        input_name=node.inputs[0],
        output_name=node.outputs[0],
        reps=node.input_tensors[node.inputs[1]].astype(np.int32).tolist(),
    )


def _convert_topk(builder, node, graph, err):
    """
    convert to CoreML TopK Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L5190
    """
    load_input_constants(builder, node, graph, err)
    axis = node.attrs.get("axis", -1)
    bottom_k = node.attrs.get("largest", True) == False
    k = node.attrs.get('k', 0)
    # NOTE: Sorted order attribute is currently ignored in CoreML
    sorted_order = node.attrs.get("sorted", True)
    if "sorted" in node.attrs:
        err.unsupported_feature_warning(
            node, "Sorted Order attribute('sorted') is currently ignored in CoreML 3.0"
        )

    builder.add_topk(
        name=node.name,
        input_names=node.inputs,
        output_names=node.outputs,
        k=k,
        axis=axis,
        use_bottom_k=bottom_k,
    )


def _convert_transpose(builder, node, graph, err):
    """
    convert to CoreML Transpose Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3426
    """

    axes = node.attrs.get("perm", [])
    # If 'perm' not provided, the reverse the dimensions
    if axes == []:
        rank = builder._get_rank(node.inputs[0])
        if rank == -1:
            return err.unsupported_op_configuration(
                builder, node, graph, "Rank unknown for input"
            )
        axes = list(range(-1, -(rank + 1), -1))

    builder.add_transpose(
        name=node.name,
        axes=axes,
        input_name=node.inputs[0],
        output_name=node.outputs[0],
    )


def _convert_unsqueeze(builder, node, graph, err):
    """
    convert to CoreML ExpandDim Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L4810
    """
    axes = node.attrs.get("axes")
    builder.add_expand_dims(
        name=node.name,
        input_name=node.inputs[0],
        output_name=node.outputs[0],
        axes=axes,
    )


def _convert_where(builder, node, graph, err):
    """
    convert to CoreML WhereBroadcastable Layer:
    https://github.com/apple/coremltools/blob/655b3be5cc0d42c3c4fa49f0f0e4a93a26b3e492/mlmodel/format/NeuralNetwork.proto#L3742
    """
    load_input_constants(builder, node, graph, err)
    builder.add_where_broadcastable(
        name=node.name, input_names=node.inputs, output_name=node.outputs[0],
    )


_ONNX_NODE_REGISTRY_ND = {
    "Abs": _convert_abs,
    "Acos": _convert_acos,
    "Acosh": _convert_acosh,
    "Add": _convert_add,
    "And": _convert_logical,
    "ArgMax": _convert_argmax,
    "ArgMin": _convert_argmin,
    "Asin": _convert_asin,
    "Asinh": _convert_asinh,
    "Atan": _convert_atan,
    "Atanh": _convert_atanh,
    "AveragePool": _convert_pool,
    "BatchNormalization": _convert_bn,
    "Cast": _convert_cast,
    "Ceil": _convert_ceil,
    "Clip": _convert_clip,
    "Concat": _convert_concat,
    "Constant": _convert_constant,
    "ConstantOfShape": _convert_constant_of_shape,
    "Conv": _convert_conv,
    "ConvTranspose": _convert_conv,
    "Cos": _convert_cos,
    "Cosh": _convert_cosh,
    "DepthToSpace": _convert_reorganize_data,
    "Div": _convert_div,
    "Elu": _convert_elu,
    "Equal": _convert_equal,
    "Erf": _convert_erf,
    "Exp": _convert_exp,
    "Expand": _convert_expand,
    "Flatten": _convert_flatten,
    "Floor": _convert_floor,
    "Gather": _convert_gather,
    "Gemm": _convert_gemm,
    "Greater": _convert_greater,
    "GRU": _convert_gru,
    "GlobalAveragePool": _convert_pool,
    "GlobalMaxPool": _convert_pool,
    "HardSigmoid": _convert_hardsigmoid,
    "Identity": _convert_identity,
    "InstanceNormalization": _convert_instancenorm,
    "LeakyRelu": _convert_leaky_relu,
    "Log": _convert_log,
    "LogSoftmax": _convert_softmax,
    "LRN": _convert_lrn,
    "Less": _convert_less,
    "LSTM": _convert_lstm,
    "MatMul": _convert_matmul,
    "Max": _convert_max,
    "MaxPool": _convert_pool,
    "Mean": _convert_mean,
    "Min": _convert_min,
    "Mod": _convert_mod,
    "Mul": _convert_mul,
    "Neg": _convert_neg,
    "NonZero": _convert_nonzero,
    "Not": _convert_logical,
    "Or": _convert_logical,
    "Pad": _convert_pad,
    "Pow": _convert_pow,
    "PRelu": _convert_prelu,
    "RandomNormal": _convert_randomnormal,
    "RandomNormalLike": _convert_randomnormallike,
    "RandomUniform": _convert_randomuniform,
    "RandomUniformLike": _convert_randomuniformlike,
    "Reciprocal": _convert_reciprocal,
    "ReduceL1": _convert_reduce,
    "ReduceL2": _convert_reduce,
    "ReduceLogSum": _convert_reduce,
    "ReduceLogSumExp": _convert_reduce,
    "ReduceMax": _convert_reduce,
    "ReduceMean": _convert_reduce,
    "ReduceMin": _convert_reduce,
    "ReduceProd": _convert_reduce,
    "ReduceSum": _convert_reduce,
    "ReduceSumSquare": _convert_reduce,
    "Relu": _convert_relu,
    "Reshape": _convert_reshape,
    "Resize": _convert_resize,
    "ReverseSequence": _convert_reverse_sequence,
    "RoiAlign": _convert_roialign,
    "Round": _convert_round,
    "Scatter": _convert_scatter,
    "Selu": _convert_selu,
    "Sigmoid": _convert_sigmoid,
    "Sign": _convert_sign,
    "Size": _convert_size,
    "Slice": _convert_slice,
    "Softmax": _convert_softmax,
    "Softplus": _convert_softplus,
    "Softsign": _convert_softsign,
    "SpaceToDepth": _convert_reorganize_data,
    "Split": _convert_split,
    "Shape": _convert_shape,
    "Sqrt": _convert_sqrt,
    "Squeeze": _convert_squeeze,
    "Sub": _convert_sub,
    "Sum": _convert_add,
    "Tanh": _convert_tanh,
    "ThresholdedRelu": _convert_thresholdedrelu,
    "Tile": _convert_tile,
    "TopK": _convert_topk,
    "Transpose": _convert_transpose,
    "Unsqueeze": _convert_unsqueeze,
    "Upsample": _convert_upsample,
    "Xor": _convert_logical,
    "Where": _convert_where,
}


def _get_node_converter_fn(
    builder, node, err
):  # type: (NeuralNetworkBuilder, Node, ErrorHandling) -> Callable[[NeuralNetworkBuilder, Node, Graph, ErrorHandling], None]
    """
    Get the right converter function for ONNX node op_type
    """
    op_type = node.op_type
    # Return custom conversion function if provided
    # If both node type and node name custom function
    # is provided, then use node name specific custom function, as
    # type specific custom function is more generic than name specific
    if node.name in err.custom_conversion_functions:
        return err.custom_conversion_functions[node.name]
    elif op_type in err.custom_conversion_functions:
        return err.custom_conversion_functions[op_type]
    elif op_type in _ONNX_NODE_REGISTRY_ND:
        return _ONNX_NODE_REGISTRY_ND[op_type]
    else:
        return err.unsupported_op(node)


def _convert_node_nd(
    builder, node, graph, err
):  # type: (NeuralNetworkBuilder, Node, Graph, ErrorHandling) -> None
    converter_fn = _get_node_converter_fn(builder, node, err)
    return converter_fn(builder, node, graph, err)



#  Copyright (c) 2020, Apple Inc. All rights reserved.
#
#  Use of this source code is governed by a BSD-3-clause license that can be
#  found in the LICENSE.txt file or at https://opensource.org/licenses/BSD-3-Clause

from coremltools.converters.mil.mil.ops.defs._utils import spatial_dimensions_out_shape

from ._op_reqs import *

"""
Pooling Op Superclass
"""


class Pooling(Operation):
    input_spec = InputSpec(
        x=TensorInputType(),
        kernel_sizes=IntTensorInputType(const=True),
        strides=IntTensorInputType(const=True, optional=True),
        pad_type=StringInputType(const=True),
        pad=IntTensorInputType(const=True, optional=True),
    )

    def default_inputs(self):
        num_spatial_dims = self.x.rank - 2
        return DefaultInputs(
            strides=[1]*num_spatial_dims,
            pad=[0]*2*num_spatial_dims,
            )

    def __init__(self, **kwargs):
        super(Pooling, self).__init__(**kwargs)

    def type_inference(self):
        ksize = self.kernel_sizes.val
        x_shape = self.x.shape
        D_in_rank = len(x_shape) - 2

        strides = [1] * D_in_rank if self.strides is None else self.strides.val
        pad_type = "valid" if self.pad_type is None else self.pad_type.val.lower()
        pad = None if self.pad is None else self.pad.val
        D_in = x_shape[2:]  # spatial dimensions
        D_out_shape = spatial_dimensions_out_shape(
            pad_type=pad_type,
            input_shape=D_in,
            kernel_shape=ksize,
            strides=strides,
            custom_pad=pad,
        )
        ret_shape = list(x_shape[:2]) + D_out_shape
        return types.tensor(self.x.dtype, tuple(ret_shape))


@register_op(doc_str="")
class avg_pool(Pooling):
    """
    Perform average pooling. Currently defined only for spatial 1-D and
    2-D cases. (Can be extended to the 3-D case easily.)
    
    Parameters
    ----------
    x: tensor<[n,C_in,\*D_in],T> (Required)
        *  ``3 <= rank <= 5``.
        *  ``D_in`` are spatial dimensions, ``1 <= len(D_in) <= 2``.
        *  ``C_in`` is the number of input channels or depth dimensions.
        *  ``n`` is the batch dimension.
    
    kernel_sizes: const tensor<[K],T> (Required)
        * The size of the window for each spatial dimension ``D_in`` of the
          input tensor.
        * ``K == len(D_in)``
    
    strides: const tensor<[S],i32> (Optional, default to all 1s)
        * Stride along each of the spatial dimensions.
        * ``S == len(D_in)``.
    
    pad_type: const str (Required)
        Must be one of the following:
        
        * ``valid``: No padding. This is equivalent to custom pad with ``pad[i] = 0, for
          all i``.
        * ``custom``: Specify custom padding in the parameter pad. note that "same"
          padding is equivalent to custom padding with
          ``pad[2*i] + pad[2*i+1] = kernel_size[i]``.
    
    pad: const<[P],i32> (Optional. Default to all 0s)
        * ``pad`` represents the number of elements to pad before and after each 
          dimension: `pad[2*i], pad[2*i+1]` are the pad size before and after spatial
          dimension ``i``.
        * ``P = 2 * len(D_in)``.
        * ``pad`` should be specified if and only if ``pad_type == custom``
    
    exclude_padding_from_average: const tensor<[], bool> (Optional, default to False)
        * If ''True'', padded values (0s) are excluded from the denominator count
          when computing the average over the kernel window.
    
    Returns
    -------
    tensor<[n, C_out,\*D_out],T>
        * Same rank as ``x``.
        * ``D_out[i] = floor[(D_in[i] + pad[2*i] + pad[2*i+1] - kernel_sizes[i]) /
          strides[i]] +1, for i = 0, .., len(D_in) - 1`` is mathematically the same
          as (when all parameters involved are integers):
          
              * ``D_out[i] = ceil [(D_in[i] + pad[2*i] + pad[2*i+1] - kernel_size[i] - 1) / stride[i]], for i = 0, .., len(D_in) - 1``.
              * ``*D_out`` is all 1s if ``global_pooling`` is ``true``.
    
    Attributes
    ----------
    T: fp32
    
    See Also
    --------
    l2_pool, max_pool
    """
    
    input_spec = (
        InputSpec(
          exclude_padding_from_average=BoolInputType(const=True,
            optional=True))
        + Pooling.input_spec
    )

    def default_inputs(self):
        num_spatial_dims = self.x.rank - 2
        return super().default_inputs() + \
            DefaultInputs(
                exclude_padding_from_average=False,
                )

    def __init__(self, **kwargs):
        super(avg_pool, self).__init__(**kwargs)


@register_op(doc_str="")
class l2_pool(Pooling):
    """
    Perform L2 pooling. Currently supports only 1-D and 2-D.
    
    Parameters
    ----------
    x: tensor<[n,C_in,*D_in],T> (Required)
        * See ``avg_pool``.
    
    kernel_sizes: const tensor<[K],T> (Required)
        * See ``avg_pool``.
    
    strides: const tensor<[S],i32> (Optional, default to all 1s)
        * See ``avg_pool``.
    
    pad_type: const str (Required)
        * See ``avg_pool``.
    
    pad: const<[P],i32> (Optional, default to all 0s)
        * See ``avg_pool``.
    
    Returns
    -------
    tensor<[n, C_out,*D_out],T>
        * See ``avg_pool``.
    
    Attributes
    ----------
    T: fp32
    
    See Also
    --------
    avg_pool, max_pool
    """
    
    def __init__(self, **kwargs):
        super(l2_pool, self).__init__(**kwargs)


@register_op(doc_str="")
class max_pool(Pooling):
    """
    Perform max pooling. Currently supports only 1-D and 2-D.
    
    Parameters
    ----------
    x: tensor<[n,C_in,*D_in],T> (Required)
        * See ``avg_pool``.
    
    kernel_sizes: const tensor<[K],T> (Required)
        * See ``avg_pool``.
    
    strides: const tensor<[S],i32> (Optional, default to all 1s)
        * See ``avg_pool``.
    
    pad_type: const str (Required)
        * See ``avg_pool``.
    
    pad: const<[P],i32> (Optional, default to all 0s)
        * See ``avg_pool``.
    
    Returns
    -------
    tensor<[n, C_out,*D_out],T>
        * See ``avg_pool``.
    
    Attributes
    ----------
    T: fp32

    See Also
    --------
    avg_pool, l2_pool
    """
    
    def __init__(self, **kwargs):
        super(max_pool, self).__init__(**kwargs)
